import torch
from torch import nn
import torch
from torch.autograd import Variable, grad

from ..distance.electron_electron_distance import ElectronElectronDistance
from ..distance.electron_nuclei_distance import ElectronNucleiDistance


class JastrowFactorElectronElectronNuclei(nn.Module):

    def __init__(self, nup, ndown, atomic_pos,
                 jastrow_kernel,
                 kernel_kwargs={},
                 cuda=False):
        """Jastrow Factor of the elec-elec-nuc term:

        .. math::
            J =  \\exp\\left(  \\sum_A \\sum_{i<j} K(R_{iA}, r_{jA}, r_{rij}) \\right)

        Args:
            nup (int): number of spin up electons
            ndow (int): number of spin down electons
            cuda (bool, optional): Turns GPU ON/OFF. Defaults to False.
        """

        super().__init__()

        self.nup = nup
        self.ndown = ndown
        self.nelec = nup + ndown

        self.cuda = cuda
        self.device = torch.device('cpu')
        if self.cuda:
            self.device = torch.device('cuda')

        self.atoms = atomic_pos.to(self.device)
        self.natoms = atomic_pos.shape[0]
        self.ndim = 3

        # kernel function
        self.jastrow_kernel = jastrow_kernel(nup, ndown,
                                             atomic_pos,
                                             cuda,
                                             **kernel_kwargs)

        # requires autograd to compute derivatives
        self.requires_autograd = self.jastrow_kernel.requires_autograd

        # index to extract tri up matrices
        self.mask_tri_up, self.index_col, self.index_row = self.get_mask_tri_up()
        self.index_elec = [
            self.index_row.tolist(), self.index_col.tolist()]

        # distance calculator
        self.elel_dist = ElectronElectronDistance(
            self.nelec, self.ndim)
        self.elnu_dist = ElectronNucleiDistance(
            self.nelec, self.atoms, self.ndim)

        # method to compute the second derivative
        # If False jastrow_factor_second_derivative will be used
        # this method  only works when the kernel does not multiply
        # the different terms e.g : k = f(r_ij) + g(R_iA) + g(R_jA)
        # For non-lienar kernels e.g. : k = f(r_ij) * g(R_iA) * g(R_jA)
        # auto_second_derivative must be set to True.
        self.auto_second_derivative = True

    def get_mask_tri_up(self):
        r"""Get the mask to select the triangular up matrix

        Returns:
            torch.tensor: mask of the tri up matrix
        """
        mask = torch.zeros(self.nelec, self.nelec).type(
            torch.bool).to(self.device)
        index_col, index_row = [], []
        for i in range(self.nelec-1):
            for j in range(i+1, self.nelec):
                index_row.append(i)
                index_col.append(j)
                mask[i, j] = True

        index_col = torch.LongTensor(index_col).to(self.device)
        index_row = torch.LongTensor(index_row).to(self.device)
        return mask, index_col, index_row

    def extract_tri_up(self, inp):
        r"""extract the upper triangular elements

        Args:
            input (torch.tensor): input matrices (..., nelec, nelec)

        Returns:
            torch.tensor: triangular up element (..., nelec_pair)
        """
        shape = list(inp.shape)
        out = inp.masked_select(self.mask_tri_up)
        return out.view(*(shape[:-2] + [-1]))

    def extract_elec_nuc_dist(self, en_dist):
        r"""Organize the elec nuc distances

        Args:
            en_dist (torch.tensor): electron-nuclei distances
                                    nbatch x nelec x natom or
                                    nbatch x 3 x nelec x natom (dr)

        Returns:
            torch.tensor: nbatch x natom x nelec_pair x 2 or
            torch.tensor: nbatch x 3 x natom x nelec_pair x 2 (dr)
        """
        out = en_dist[..., self.index_elec, :]
        if en_dist.ndim == 3:
            return out.permute(0, 3, 2, 1)
        elif en_dist.ndim == 4:
            return out.permute(0, 1, 4, 3, 2)
        else:
            raise ValueError(
                'elec-nuc distance matrix should have 3 or 4 dim')

    def assemble_dist(self, pos):
        """Assemle the different distances for easy calculations

        Args:
            pos (torch.tensor): Positions of the electrons
                                  Size : Nbatch, Nelec x Ndim

        Returns:
            torch.tensor : nbatch, natom, nelec_pair, 3

        """

        # get the elec-elec distance matrix
        ree = self.extract_tri_up(self.elel_dist(pos))
        ree = ree.unsqueeze(1).unsqueeze(-1)
        ree = ree.repeat(1, self.natoms, 1, 1)

        # get the elec-nuc distance matrix
        ren = self.extract_elec_nuc_dist(self.elnu_dist(pos))

        # cat both
        return torch.cat((ren, ree), -1)

    def assemble_dist_deriv(self, pos, derivative=1):
        """Assemle the different distances for easy calculations
           the output has dimension  nbatch, 3 x natom, nelec_pair, 3
           the last dimension is composed of [r_{e_1n}, r_{e_2n}, r_{ee}]

        Args:
            pos (torch.tensor): Positions of the electrons
                                  Size : Nbatch, Nelec x Ndim

        Returns:
            torch.tensor : nbatch, 3 x natom, nelec_pair, 3

        """

        # get the elec-elec distance derivative
        dree = self.elel_dist(pos, derivative)
        dree = self.extract_tri_up(dree)
        dree = dree.unsqueeze(2).unsqueeze(-1)
        dree = dree.repeat(1, 1, self.natoms, 1, 1)

        # get the elec-nuc distance derivative
        dren = self.elnu_dist(pos, derivative)
        dren = self.extract_elec_nuc_dist(dren)

        # assemble
        return torch.cat((dren, dree), -1)

    def _to_device(self):
        """Export the non parameter variable to the device."""

        self.device = torch.device('cuda')
        self.to(self.device)
        attrs = ['static_weight']
        for at in attrs:
            if at in self.__dict__:
                self.__dict__[at] = self.__dict__[at].to(self.device)

    def forward(self, pos, derivative=0, sum_grad=True):
        """Compute the Jastrow factors.

        Args:
            pos (torch.tensor): Positions of the electrons
                                  Size : Nbatch, Nelec x Ndim
            derivative (int, optional): order of the derivative (0,1,2,).
                            Defaults to 0.
            sum_grad (bool, optional): Return the sum_grad (i.e. the sum of
                                       the derivatives) or the individual
                                       terms. Defaults to True.
                                       False only for derivative=1

        Returns:
            torch.tensor: value of the jastrow parameter for all confs
                          derivative = 0  (Nmo) x Nbatch x 1
                          derivative = 1  (Nmo) x Nbatch x Nelec (for sum_grad = True)
                          derivative = 1  (Nmo) x Nbatch x Ndim x Nelec (for sum_grad = False)
                          derivative = 2  (Nmo) x Nbatch x Nelec
        """

        size = pos.shape
        assert size[1] == self.nelec * self.ndim
        nbatch = size[0]

        r = self.assemble_dist(pos)
        kern_vals = self.jastrow_kernel(r)
        jast = torch.exp(kern_vals.view(nbatch, -1).sum(-1))

        if derivative == 0:
            return jast.unsqueeze(-1)

        elif derivative == 1:
            dr = self.assemble_dist_deriv(pos, 1)
            return self.jastrow_factor_derivative(r, dr, jast, sum_grad)

        elif derivative == 2:

            if self.auto_second_derivative:
                return self.jastrow_factor_second_derivative_auto(pos, jast=jast.unsqueeze(-1))

            else:
                dr = self.assemble_dist_deriv(pos, 1)
                d2r = self.assemble_dist_deriv(pos, 2)

                return self.jastrow_factor_second_derivative(r, dr, d2r, jast)

        elif derivative == [0, 1, 2]:

            dr = self.assemble_dist_deriv(pos, 1)
            djast = self.jastrow_factor_derivative(
                r, dr, jast, sum_grad)

            if self.auto_second_derivative:
                d2jast = self.jastrow_factor_second_derivative_auto(
                    pos, jast=jast.unsqueeze(-1))
            else:
                d2r = self.assemble_dist_deriv(pos, 2)
                d2jast = self.jastrow_factor_second_derivative(
                    r, dr, d2r, jast)

            return(jast.unsqueeze(-1), djast, d2jast)

        else:
            raise ValueError('Derivative value nor recognized')

    def jastrow_factor_derivative(self, r, dr, jast, sum_grad):
        """Compute the value of the derivative of the Jastrow factor

        Args:
            r (torch.tensor): ee distance matrix Nbatch x Nelec x Nelec
            jast (torch.tensor): values of the jastrow elements
                                 Nbatch x Nelec x Nelec

        Returns:
            torch.tensor: gradient of the jastrow factors
                          Nbatch x Nelec x Ndim
        """

        if sum_grad:

            # derivative of the jastrow elements
            # nbatch x ndim x natom x nelec_pair x 3
            # last dim is (ria rja rij)
            djast = self.jastrow_kernel.compute_derivative(r, dr)

            # sum dim and atom
            djast = djast.sum([1, 2])

            # multiply with the product of jastrow el values
            djast = djast * jast.unsqueeze(-1).unsqueeze(-1)

            # create the output vector with size nbatch x nelec
            out_shape = list(djast.shape[:-2]) + [self.nelec]
            out = torch.zeros(out_shape).to(self.device)

            # add the elec-elec term
            out.index_add_(-1, self.index_row, djast[..., 2])
            out.index_add_(-1, self.index_col, -djast[..., 2])

            # add the elec-nuc terms
            out.index_add_(-1, self.index_row, djast[..., 0])
            out.index_add_(-1, self.index_col, djast[..., 1])

        else:

            # derivative of the jastrow elements
            # nbatch x ndim x natom x nelec_pair x 3
            # last dim is (ria rja rij)
            djast = self.jastrow_kernel.compute_derivative(r, dr)

            # sum atom
            djast = djast.sum(2)
            djast = djast * \
                jast.unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)

            # might cause problems with backward cause in place operation
            out_shape = list(djast.shape[:-2]) + [self.nelec]
            out = torch.zeros(out_shape).to(self.device)

            # add electronic terms
            out.index_add_(-1, self.index_row, djast[..., 2])
            out.index_add_(-1, self.index_col, -djast[..., 2])

            # add elec-nuc terms
            out.index_add_(-1, self.index_row, djast[..., 0])
            out.index_add_(-1, self.index_col, djast[..., 1])

        return out

    def jastrow_factor_second_derivative(self, r, dr, d2r, jast):
        """Compute the value of the pure 2nd derivative of the Jastrow factor

        Args:
            r (torch.tensor): ee distance matrix Nbatch x Nelec x Nelec
            jast (torch.tensor): values of the ajstrow elements
                                 Nbatch x Nelec x Nelec

        Returns:
            torch.tensor: diagonal hessian of the jastrow factors
                          Nbatch x Nelec x Ndim
        """

        # puresecond derivative of the jast el
        # nbatch x ndim x natom x nelec_pair x 3
        # last dim is (ria rja rij)
        d2jast = self.jastrow_kernel.compute_second_derivative(
            r, dr, d2r)

        # sum over the dim and the atom
        d2jast = d2jast.sum([1, 2])

        # might cause problems with backward cause in place operation
        hess_shape = list(d2jast.shape[:-2]) + [self.nelec]
        hess_jast = torch.zeros(hess_shape).to(self.device)

        # add elec-elec terms
        hess_jast.index_add_(-1, self.index_row, d2jast[..., 2])
        hess_jast.index_add_(-1, self.index_col, d2jast[..., 2])

        # add elec-nu terms
        hess_jast.index_add_(-1, self.index_row, d2jast[..., 0])
        hess_jast.index_add_(-1, self.index_col, d2jast[..., 1])

        # mixed terms
        djast = self.jastrow_kernel.compute_derivative(r, dr)

        # add partial derivative
        hess_jast = hess_jast + self.partial_derivative(djast)

        return hess_jast * jast.unsqueeze(-1)

    def partial_derivative(self, djast):
        """[summary]

        Args:
            djast ([type]): [description]
        """

        # create the output vector with size nbatch x nelec
        out_shape = list(djast.shape[:-2]) + [self.nelec]
        out = torch.zeros(out_shape).to(self.device)

        # add the elec-elec term
        out.index_add_(-1, self.index_row, djast[..., 2])
        out.index_add_(-1, self.index_col, -djast[..., 2])

        # add the elec-nuc terms
        out.index_add_(-1, self.index_row, djast[..., 0])
        out.index_add_(-1, self.index_col, djast[..., 1])

        return ((out.sum(2))**2).sum(1)

    def jastrow_factor_second_derivative_auto(self, pos, jast=None):
        """Compute the second derivative of the jastrow factor automatically.
        This is needed for complicate kernels where the partial derivatives of
        the kernels are difficult to organize in a total derivaitve e.e Boys-Handy

        Args:
            pos ([type]): [description]
        """

        def hess(out, pos):

            # compute the jacobian
            z = Variable(torch.ones_like(out))
            jacob = grad(out, pos,
                         grad_outputs=z,
                         only_inputs=True,
                         create_graph=True)[0]

            # compute the diagonal element of the Hessian
            z = Variable(torch.ones(jacob.shape[0])).to(self.device)
            hess = torch.zeros_like(jacob)

            for idim in range(jacob.shape[1]):

                tmp = grad(jacob[:, idim], pos,
                           grad_outputs=z,
                           only_inputs=True,
                           create_graph=True)[0]

                hess[:, idim] = tmp[:, idim]

            return hess

        nbatch = pos.shape[0]
        if jast is None:
            jast = self.forward(pos)
        return hess(jast, pos).view(nbatch, self.nelec, 3).sum(2)



from .backflow_kernel_base import BackFlowKernelBase
from .backflow_kernel_autodiff_inverse import BackFlowKernelAutoInverse
from .backflow_kernel_fully_connected import BackFlowKernelFullyConnected
from .backflow_kernel_inverse import BackFlowKernelInverse
from .backflow_kernel_power_sum import BackFlowKernelPowerSum
from .backflow_kernel_square import BackFlowKernelSquare



import torch

from .. import log
from .walkers import Walkers


class SamplerBase:

    def __init__(self, nwalkers, nstep, step_size,
                 ntherm, ndecor, nelec, ndim, init,
                 cuda):
        """Base class for the sampler

        Args:
            nwalkers (int): number of walkers
            nstep (int): number of MC steps
            step_size (float): size of the steps in bohr
            ntherm (int): number of MC steps to thermalize
            ndecor (int): unmber of MC steps to decorellate
            nelec (int): number of electrons in the system
            ndim (int): number of cartesian dimension
            init (dict): method to initialize the walkers
            cuda ([type]): [description]
        """

        self.nwalkers = nwalkers
        self.nelec = nelec
        self.ndim = ndim
        self.nstep = nstep
        self.step_size = step_size
        self.ntherm = ntherm
        self.ndecor = ndecor
        self.cuda = cuda
        if cuda:
            self.device = torch.device('cuda')
        else:
            self.device = torch.device('cpu')

        self.walkers = Walkers(
            nwalkers=nwalkers, nelec=nelec, ndim=ndim, init=init, cuda=cuda)

        log.info('')
        log.info(' Monte-Carlo Sampler')
        log.info('  Number of walkers   : {0}', self.nwalkers)
        log.info('  Number of steps     : {0}', self.nstep)
        log.info('  Step size           : {0}', self.step_size)
        log.info('  Thermalization steps: {0}', self.ntherm)
        log.info('  Decorelation steps  : {0}', self.ndecor)
        log.info('  Walkers init pos    : {0}', init['method'])

    def __call__(self, pdf, *args, **kwargs):
        raise NotImplementedError(
            "Sampler must have a __call__ method")

    def __repr__(self):
        return self.__class__.__name__ + ' sampler with  %d walkers' % self.nwalkers

    def get_sampling_size(self):
        """evaluate the number of sampling point we'll have."""
        if self.ntherm == -1:
            return self.nwalkers
        else:
            return self.walkers.nwalkers * int((self.nstep-self.ntherm)/self.ndecor)



__all__ = ['CalculatorBase', 'CalculatorADF', 'CalculatorADF2019', 'CalculatorPySCF']

from .calculator_base import CalculatorBase
from .adf import CalculatorADF, CalculatorADF2019
from .pyscf import CalculatorPySCF



import torch
from torch import nn

from .jastrow_kernel_electron_nuclei_base import JastrowKernelElectronNucleiBase


class FullyConnectedJastrowKernel(JastrowKernelElectronNucleiBase):

    def __init__(self, nup, ndown, atomic_pos, cuda, w=1.):
        r"""Computes the Simple Pade-Jastrow factor

        .. math::
            J = \prod_{i<j} \exp(B_{ij}) \quad \quad \\text{with} \quad \quad
            B_{ij} = \\frac{w_0 r_{i,j}}{1 + w r_{i,j}}

        Args:
            nup (int): number of spin up electons
            ndow (int): number of spin down electons
            atoms (torch.tensor): atomic positions of the atoms
            w (float, optional): Value of the variational parameter. Defaults to 1..
            cuda (bool, optional): Turns GPU ON/OFF. Defaults to False.
        """

        super().__init__(nup, ndown, atomic_pos, cuda)

        self.fc1 = nn.Linear(1, 16, bias=False)
        self.fc2 = nn.Linear(16, 8, bias=False)
        self.fc3 = nn.Linear(8, 1, bias=False)

        self.nl_func = torch.nn.Sigmoid()
        self.requires_autograd = True

    def forward(self, x):
        """ Get the jastrow kernel.

        Args:
            x (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nnuc

        Returns:
            torch.tensor: matrix of the jastrow kernels
                          Nbatch x Nelec x Nnuc
        """
        original_shape = x.shape

        # reshape the input so that all elements are considered
        # independently of each other
        x = x.reshape(-1, 1)

        x = self.fc1(x)
        x = self.nl_func(x)
        x = self.fc2(x)
        x = self.nl_func(x)
        x = self.fc3(x)
        x = self.nl_func(x)

        # reshape to the original shape
        x = x.reshape(*original_shape)

        return x



import unittest


from qmctorch.sampler import Metropolis
from .test_sampler_base import TestSamplerBase


class TestMetropolis(TestSamplerBase):

    def test_metropolis(self):
        """Test Metropolis sampling."""

        sampler = Metropolis(
            nwalkers=10,
            nstep=20,
            step_size=0.5,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('normal'))

        for m in ['one-elec', 'all-elec', 'all-elec-iter']:
            for p in ['normal', 'uniform']:

                sampler.configure_move({'type': m, 'proba': p})
                pos = sampler(self.wf.pdf)


if __name__ == "__main__":
    unittest.main()




import torch
from torch import nn
from torch.autograd import grad
from .kernels.jastrow_kernel_electron_electron_base import JastrowKernelElectronElectronBase


class OrbitalDependentJastrowKernel(JastrowKernelElectronElectronBase):

    def __init__(self, nup, ndown, nmo, cuda,
                 jastrow_kernel, kernel_kwargs={}):
        """Transform a kernel into a orbital dependent kernel

        Args:
            nup (int): number of spin up electrons
            ndown (int): number of spin down electron
            nmo (int): number of orbital
            cuda (bool): use GPUs
            jastrow_kernel (kernel function): kernel to be used
            kernel_kwargs (dict): keyword arguments of the kernel
        """

        super().__init__(nup, ndown, cuda)
        self.nmo = nmo
        self.jastrow_functions = nn.ModuleList(
            [jastrow_kernel(nup, ndown, cuda, **kernel_kwargs) for _ in range(self.nmo)])

    def forward(self, r):
        """ Get the jastrow kernel.

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec_pair

        Returns:
            torch.tensor: matrix of the jastrow kernels
                          Nmo x Nbatch x Nelec_pair
        """
        out = None
        for jast in self.jastrow_functions:
            jvals = jast(r).unsqueeze(0)
            if out is None:
                out = jvals
            else:
                out = torch.cat((out, jvals), axis=0)
        return out

    def compute_derivative(self, r, dr):
        """Get the elements of the derivative of the jastrow kernels
        wrt to the first electrons

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec_pair
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec_pair

        Returns:
            torch.tensor: matrix fof the derivative of the jastrow elements
                          Nmo x Nbatch x Ndim x  Nelec_pair
        """

        out = None

        if r.requires_grad == False:
            r.requires_grad = True

        with torch.enable_grad():

            for jast in self.jastrow_functions:

                kernel = jast(r)
                ker_grad = self._grads(kernel, r)
                ker_grad = ker_grad.unsqueeze(1) * dr
                ker_grad = ker_grad.unsqueeze(0).detach().clone()

                if out is None:
                    out = ker_grad
                else:
                    out = torch.cat((out, ker_grad), axis=0)

        return out

    def compute_second_derivative(self, r, dr, d2r):
        """Get the elements of the pure 2nd derivative of the jastrow kernels
        wrt to the first electron


        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec_pair
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec_pair
            d2r (torch.tensor): matrix of the 2nd derivative of
                                the e-e distances
                              Nbatch x Ndim x Nelec_pair

        Returns:
            torch.tensor: matrix fof the pure 2nd derivative of
                          the jastrow elements
                          Nmo x Nbatch x Ndim x Nelec_pair
        """
        dr2 = dr * dr
        out = None

        if r.requires_grad == False:
            r.requires_grad = True

        with torch.enable_grad():

            for jast in self.jastrow_functions:

                kernel = jast(r)
                ker_hess, ker_grad = self._hess(kernel, r)

                jhess = (ker_hess).unsqueeze(1) * \
                    dr2 + ker_grad.unsqueeze(1) * d2r

                jhess = jhess.unsqueeze(0)

                if out is None:
                    out = jhess
                else:
                    out = torch.cat((out, jhess))

        return out

    @staticmethod
    def _grads(val, r):
        """Get the gradients of the jastrow values
        of a given orbital terms

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec_pair

        Returns:
            torch.tensor: gradients of the values wrt to ee distance
        """
        return grad(val, r, grad_outputs=torch.ones_like(val))[0]

    @staticmethod
    def _hess(val, r):
        """get the hessian of the jastrow values.
        of a given orbital terms
        Warning thos work only because the orbital term are dependent
        of a single rij term, i.e. fij = f(rij)


        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec_pair

        Returns:
            torch.tensor: second derivative of the values wrt to ee distance
        """

        gval = grad(val, r,
                    grad_outputs=torch.ones_like(val),
                    create_graph=True)[0]

        hval = grad(gval, r,
                    grad_outputs=torch.ones_like(gval))[0]

        return hval, gval



import unittest
from qmctorch.wavefunction.orbitals.radial_functions import (radial_gaussian_pure,
                                                             radial_gaussian,
                                                             radial_slater,
                                                             radial_slater_pure)
import torch
from torch.autograd import grad, Variable


def hess(out, pos):

    # compute the jacobian
    z = Variable(torch.ones(out.shape))
    jacob = grad(out, pos,
                 grad_outputs=z,
                 only_inputs=True,
                 create_graph=True)[0]

    # compute the diagonal element of the Hessian
    z = Variable(torch.ones(jacob.shape[0]))
    hess = torch.zeros(jacob.shape)

    for idim in range(jacob.shape[1]):

        tmp = grad(jacob[:, idim], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, idim] = tmp[:, idim]

    return hess


def hess_mixed_terms(out, pos):

    # compute the jacobian
    z = Variable(torch.ones(out.shape))
    jacob = grad(out, pos,
                 grad_outputs=z,
                 only_inputs=True,
                 create_graph=True)[0]

    # compute the diagonal element of the Hessian
    z = Variable(torch.ones(jacob.shape[0]))
    hess = torch.zeros(jacob.shape)
    nelec = pos.shape[1]//3
    k = 0

    for ielec in range(nelec):

        ix = ielec*3
        tmp = grad(jacob[:, ix], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, k] = tmp[:, ix+1]
        k = k + 1
        hess[:, k] = tmp[:, ix+2]
        k = k + 1

        iy = ielec*3 + 1
        tmp = grad(jacob[:, iy], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, k] = tmp[:, iy+1]
        k = k + 1

    return hess


class TestRadialFunctions(unittest.TestCase):

    def setUp(self):
        self.radfn = [radial_gaussian,
                      radial_gaussian_pure,
                      radial_slater,
                      radial_slater_pure]

        self.nbatch = 10
        self.nelec = 4
        self.nbas = 6

        self.bas_n = torch.Tensor([0, 1, 1, 1, 2, 2])
        self.bas_exp = torch.rand(self.nbas)

        self.xyz = Variable(torch.rand(self.nbatch, self.nelec*3))
        self.xyz.requires_grad = True

    def process_position(self):
        """Return the distance between electron and centers."""

        bas_coords = torch.zeros(self.nbas, 3)
        xyz = (self.xyz.view(-1, self.nelec, 1, 3) -
               bas_coords[None, ...])
        r = torch.sqrt((xyz*xyz).sum(3))
        return xyz, r

    def test_val(self):
        """Simply executes the kernel."""
        xyz, r = self.process_position()
        for fn in self.radfn:
            fn(r, self.bas_n, self.bas_exp)

    def test_grad(self):
        """Compute the gradients of the radial function."""

        for fn in self.radfn:

            xyz, r = self.process_position()
            val = fn(r, self.bas_n, self.bas_exp)
            val_grad_auto = grad(
                val, self.xyz, torch.ones_like(val))[0]

            val_grad = fn(r, self.bas_n, self.bas_exp, xyz=xyz,
                          derivative=1, sum_grad=False)

            val_grad_sum = fn(r, self.bas_n, self.bas_exp, xyz=xyz,
                              derivative=1, sum_grad=True)

            assert(torch.allclose(
                val_grad.sum(), val_grad_auto.sum(), atol=1E-6))

            assert(torch.allclose(
                val_grad.sum(-1), val_grad_sum, atol=1E-6))

    def test_lap(self):
        """Computes the laplacian of the radial functions."""

        for fn in self.radfn:

            xyz, r = self.process_position()
            val = fn(r, self.bas_n, self.bas_exp)

            val_lap = fn(r, self.bas_n, self.bas_exp, xyz=xyz,
                         derivative=2, sum_hess=False)
            val_lap_sum = fn(r, self.bas_n, self.bas_exp, xyz=xyz,
                             derivative=2, sum_hess=True)

            val_lap_auto = hess(val, self.xyz)

            assert(torch.allclose(
                val_lap.sum(-1), val_lap_sum, atol=1E-6))

            assert(torch.allclose(
                val_lap.sum(), val_lap_auto.sum(), atol=1E-6))

    def test_mixed(self):
        """Test the mixed second derivatives."""
        for fn in self.radfn:
            xyz, r = self.process_position()
            val = fn(r, self.bas_n, self.bas_exp)
            val_lap = fn(r, self.bas_n, self.bas_exp, xyz=xyz,
                         derivative=3)

        val_lap_auto = hess_mixed_terms(val, self.xyz)
        assert(torch.allclose(
            val_lap.sum(), val_lap_auto.sum(), atol=1E-6))


if __name__ == "__main__":
    # unittest.TestCase()
    t = TestRadialFunctions()
    t.setUp()
    t.test_grad()
    t.test_lap()
    t.test_mixed()



import torch
from torch import nn
from torch.autograd import grad, Variable
from .backflow_kernel_base import BackFlowKernelBase


class BackFlowKernelFullyConnected(BackFlowKernelBase):

    def __init__(self, mol, cuda):
        """Compute the back flow kernel, i.e. the function
        f(rij) where rij is the distance between electron i and j
        This kernel is used in the backflow transformation
        .. math:
            q_i = r_i + \\sum_{j\\neq i} f(r_{ij}) (r_i-r_j)
        """
        super().__init__(mol, cuda)
        self.fc1 = nn.Linear(1, 16, bias=False)
        self.fc2 = nn.Linear(16, 1, bias=False)
        self.nl_func = torch.nn.Sigmoid()

        eps = 1E-0
        self.fc1.weight.data *= eps
        self.fc2.weight.data *= eps

    def _backflow_kernel(self, ree):
        """Computes the kernel via autodiff

        Args:
            ree ([type]): [description]

        Returns:
            [type]: [description]
        """
        original_shape = ree.shape
        x = ree.reshape(-1, 1)

        x = self.fc1(x)
        x = self.nl_func(x)
        x = self.fc2(x)
        x = self.nl_func(x)
        x = x.reshape(*original_shape)

        return x



import torch
from torch import nn
from .jastrow_kernel_electron_electron_nuclei_base import JastrowKernelElectronElectronNucleiBase


class BoysHandyJastrowKernel(JastrowKernelElectronElectronNucleiBase):

    def __init__(self, nup, ndown, atomic_pos, cuda, nterm=5):
        """Defines a Boys Handy jastrow factors.

        J.W. Moskowitz et. al
        Correlated Monte Carlo Wave Functions for Some Cations and Anions of the First Row Atoms
        Journal of Chemical Physics, 97, 3382-85 (1992)

        .. math::
            \\text{K}(R_{iA}, R_{jA}, r_{ij) = \\sum_\\mu c_\\mu \\left(\\frac{a_{1_\\mu} R_{iA}}{1 + b_{1_\\mu}R_{iA}}\\right)^{u_\\mu}
                                                                 \\left(\\frac{a_{2_\\mu} R_{jA}}{1 + b_{2_\\mu}R_{iA}}\\right)^{v_\\mu}
                                                                 \\left(\\frac{a_{3_\\mu} r_{ij}}{1 + b_{3_\\mu}r_{ij}}\\right)^{w_\\mu}

        We restrict the parameters of the two electron-nucleus distance to be equal
        otherwise the jastrow factor is not permutation invariant

        """

        super().__init__(nup, ndown, atomic_pos, cuda)

        self.nterm = nterm
        self.fc = nn.Linear(self.nterm, 1, bias=False)

        self.weight_num = nn.Parameter(torch.rand(1, 2, self.nterm))
        self.weight_denom = nn.Parameter(torch.rand(1, 2, self.nterm))
        self.exp = nn.Parameter(torch.ones(2, self.nterm))
        self.repeat_dim = torch.as_tensor([2, 1]).to(self.device)

    def forward(self, x):
        """Compute the values of the kernel

        Args:
            x (torch.tensor): e-e and e-n distances distance (Nbatch, Natom, Nelec_pairs, 3)
                              the last dimension holds the values [R_{iA}, R_{jA}, r_{ij}]
                              where i,j are electron index in the pair and A the atom index.

        Returns:
            torch.tensor: values of the kernel (Nbatch, Natom, Nelec_pairs, 1)

        """

        # return (1E-5*x**3).prod(-1, keepdim=True)

        # reshape the input so that all elements
        # are considered independently of each other
        out_shape = list(x.shape)[:-1] + [1]

        # reshape to [N, 3 ,1]
        # here N is the
        x = x.reshape(-1, 3, 1)

        # compute the different terms :
        # x[0] = (a r_{iA})/(1 + b r_{iA})
        # x[1] = (a r_{jA})/(1 + b r_{jA})
        # x[2] = (a r_{ij})/(1 + b r_{ij})
        # output shape : [N, 3, nterm]
        wnum = self.weight_num.repeat_interleave(
            self.repeat_dim, dim=1)
        wdenom = self.weight_denom.repeat_interleave(
            self.repeat_dim, dim=1)
        x = (wnum * x) / (1. + wdenom * x)

        # comput the powers
        xp = self.exp.repeat_interleave(self.repeat_dim, dim=0)
        x = x**(xp)

        # product over the r_{iA}, r_{jA}, r_{ij}
        # output shape : [N, nterm]
        x = x.prod(1)

        # compute the sum over the different terms
        # output shape :  [N]
        x = self.fc(x)

        # print(x.reshape(*out_shape))
        return x.reshape(*out_shape)



import os
import shutil
import warnings
from types import SimpleNamespace

import numpy as np

from ... import log
from .calculator_base import CalculatorBase

try:
    from scm import plams
except ModuleNotFoundError:
    warnings.warn('scm python module not found')


class CalculatorADF(CalculatorBase):

    def __init__(self, atoms, atom_coords, basis, scf, units, molname, savefile):

        CalculatorBase.__init__(
            self, atoms, atom_coords, basis, scf, units, molname, 'adf', savefile)

        # basis from the emma paper
        self.additional_basis_type = ['VB1', 'VB2', 'VB3',
                                      'CVB1', 'CVB2', 'CVB3']

        self.additional_basis_path = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), 'atomicdata/adf/')

        self.adf_version = 'adf2020+'
        self.job_name = ''.join(self.atoms) + '_' + self.basis_name
        self.output_file = 'adf.rkf'

    def run(self):
        """Run the calculation using ADF."""

        # path needed for the calculation
        plams_wd = './plams_workdir'
        outputdir_path = os.path.join(
            plams_wd, os.path.join(self.job_name, self.output_file))

        # get the correct exec
        plams_job = {
            'adf2020+': plams.AMSJob,
            'adf2019' : plams.ADFJob
        }[self.adf_version]

        # configure plams and run the calculation
        self.init_plams()
        mol = self.get_plams_molecule()
        sett = self.get_plams_settings()
        job = plams_job(molecule=mol, settings=sett, name=self.job_name)
        job.run()

        # extract the data to hdf5
        basis = self.get_basis_data(outputdir_path)

        # remove adf data
        if self.savefile:
            shutil.copyfile(outputdir_path, self.output_file)
            self.savefile = self.output_file
        shutil.rmtree(plams_wd)

        return basis

    def init_plams(self):
        """Init PLAMS."""
        plams.init()
        plams.config.log.stdout = -1
        plams.config.erase_workdir = True

    def get_plams_molecule(self):
        """Returns a plams molecule object."""
        mol = plams.Molecule()
        bohr2angs = 0.529177
        scale = 1.
        if self.units == 'bohr':
            scale = bohr2angs
        for at, xyz in zip(self.atoms, self.atom_coords):
            xyz = list(scale * np.array(xyz))
            mol.add_atom(plams.Atom(symbol=at, coords=tuple(xyz)))
        return mol

    def get_plams_settings(self):
        """Returns a plams setting object."""

        sett = plams.Settings()
        sett.input.ams.Task = 'SinglePoint'
        
        if self.basis_name.upper() in self.additional_basis_type:
            sett.input.adf.basis.type = 'DZP'
            parsed_atoms = []
            for at in self.atoms:
                if at not in parsed_atoms:
                    basis_path = os.path.join(self.additional_basis_path, self.basis_name.upper(), at)
                    atomtype = f"Symbol={at} File={basis_path}"
                    sett.input.adf.basis.peratomtype = atomtype
                    parsed_atoms.append(at)
        else:
            sett.input.adf.basis.type = self.basis_name.upper()

        sett.input.adf.basis.core = 'None'
        sett.input.adf.symmetry = 'nosym'

        if self.scf.lower() == 'hf':
            sett.input.adf.XC.HartreeFock = ''

        elif self.scf.lower() == 'dft':
            sett.input.adf.XC.LDA = 'VWN'

        sett.input.adf.relativity.level = 'None'

        # total energy
        sett.input.adf.totalenergy = True

        return sett

    def get_basis_data(self, kffile):
        """Save the basis information needed to compute the AO values."""

        if not os.path.isfile(kffile):
            raise FileNotFoundError(
                'File %s not found, ADF may have crashed, look into the plams_workdir directory' % kffile)
        kf = plams.KFFile(kffile)
        status = kf.read('General', 'termination status').strip()
        if status != 'NORMAL TERMINATION':
            log.info(
                '  WARNING : ADF calculation terminated with status')
            log.info('          : %s' % status)
            log.info('          : Proceed with caution')

        basis = SimpleNamespace()

        basis.TotalEnergy = kf.read('Total Energy', 'Total energy')
        basis.radial_type = 'sto'
        basis.harmonics_type = 'cart'

        nao = kf.read('Basis', 'naos')
        nmo = kf.read('A', 'nmo_A')
        basis.nao = nao
        basis.nmo = nmo

        # number of bas per atom type
        nbptr = kf.read('Basis', 'nbptr')

        # number of atom per atom typ
        nqptr = kf.read('Geometry', 'nqptr')
        atom_type = kf.read('Geometry', 'atomtype').split()

        # number of bas per atom type
        nshells = np.array([nbptr[i] - nbptr[i - 1]
                            for i in range(1, len(nbptr))])

        # kx/ky/kz/kr exponent per atom type
        bas_kx = self.read_array(kf, 'Basis', 'kx')
        bas_ky = self.read_array(kf, 'Basis', 'ky')
        bas_kz = self.read_array(kf, 'Basis', 'kz')
        bas_kr = self.read_array(kf, 'Basis', 'kr')

        # bas exp/coeff/norm per atom type
        bas_exp = self.read_array(kf, 'Basis', 'alf')
        bas_norm = self.read_array(kf, 'Basis', 'bnorm')

        basis_nshells = []
        basis_bas_kx, basis_bas_ky, basis_bas_kz = [], [], []
        basis_bas_kr = []
        basis_bas_exp, basis_bas_norm = [], []

        for iat, at in enumerate(atom_type):

            number_copy = nqptr[iat + 1] - nqptr[iat]
            idx_bos = list(range(nbptr[iat] - 1, nbptr[iat + 1] - 1))

            basis_nshells += [nshells[iat]] * number_copy

            basis_bas_kx += list(bas_kx[idx_bos]) * number_copy
            basis_bas_ky += list(bas_ky[idx_bos]) * number_copy
            basis_bas_kz += list(bas_kz[idx_bos]) * number_copy
            basis_bas_kr += list(bas_kr[idx_bos]) * number_copy
            basis_bas_exp += list(bas_exp[idx_bos]) * number_copy
            basis_bas_norm += list(
                bas_norm[idx_bos]) * number_copy

        basis.nshells = basis_nshells
        basis.nao_per_atom = basis_nshells
        basis.index_ctr = np.arange(nao)
        basis.nctr_per_ao = np.ones(nao)

        basis.bas_kx = np.array(basis_bas_kx)
        basis.bas_ky = np.array(basis_bas_ky)
        basis.bas_kz = np.array(basis_bas_kz)
        basis.bas_kr = np.array(basis_bas_kr)

        basis.bas_exp = np.array(basis_bas_exp)
        basis.bas_coeffs = np.ones_like(basis_bas_exp)
        basis.bas_norm = np.array(basis_bas_norm)

        basis.atom_coords_internal = np.array(
            kf.read('Geometry', 'xyz')).reshape(-1, 3)

        # Molecular orbitals
        mos = np.array(kf.read('A', 'Eigen-Bas_A'))
        mos = mos.reshape(nmo, nao).T

        # normalize the MO
        # this is not needed !!!
        # mos = self.normalize_columns(mos)

        # orbital that take part in the rep
        npart = np.array(kf.read('A', 'npart'))-1

        # create permutation matrix
        perm_mat = np.zeros((basis.nao, basis.nao))
        for i in range(basis.nao):
            perm_mat[npart[i], i] = 1.

        # reorder the basis function
        basis.mos = perm_mat @ mos

        return basis

    @staticmethod
    def read_array(kf, section, name):
        """read a data from the kf file

        Args:
            kf (file handle): kf file
            section (str): name of the section
            name (str): name of the property

        Returns:
            np.data: data
        """
        data = np.array(kf.read(section, name))
        if data.shape == ():
            data = np.array([data])
        return data
        
class CalculatorADF2019(CalculatorADF):

    def __init__(self, atoms, atom_coords, basis, scf, units, molname, savefile):

        CalculatorADF.__init__(
            self, atoms, atom_coords, basis, scf, units, molname, savefile)

        self.adf_version = 'adf2019'
        self.job_name = ''.join(self.atoms) + '_' + self.basis_name
        self.output_file = self.job_name + '.t21'

    def get_plams_molecule(self):
        """Returns a plams molecule object."""
        mol = plams.Molecule()
        for at, xyz in zip(self.atoms, self.atom_coords):
            mol.add_atom(plams.Atom(symbol=at, coords=tuple(xyz)))
        return mol

    def get_plams_settings(self):
        """Returns a plams setting object."""

        sett = plams.Settings()
        sett.input.basis.type = self.basis_name.upper()
        if self.basis_name.upper() in self.additional_basis_type:
            sett.input.basis.path = self.additional_basis_path
        sett.input.basis.core = 'None'
        sett.input.symmetry = 'nosym'

        if self.scf.lower() == 'hf':
            sett.input.XC.HartreeFock = ''

        elif self.scf.lower() == 'dft':
            sett.input.XC.LDA = 'VWN'

        # correct unit
        if self.units == 'angs':
            sett.input.units.length = 'Angstrom'
        elif self.units == 'bohr':
            sett.input.units.length = 'Bohr'

        # total energy
        sett.input.totalenergy = True

        return sett



import unittest

import numpy as np
import torch
import torch.optim as optim
import horovod.torch as hvd
from mpi4py import MPI

from qmctorch.sampler import Metropolis
from qmctorch.solver import SolverMPI
from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow
from qmctorch.utils import set_torch_double_precision


class TestH2Hvd(unittest.TestCase):

    def setUp(self):
        hvd.init()

        torch.manual_seed(0)
        np.random.seed(0)

        set_torch_double_precision()

        # optimal parameters
        self.opt_r = 0.69  # the two h are at +0.69 and -0.69
        self.opt_sigma = 1.24

        # molecule
        self.mol = Molecule(
            atom='H 0 0 -0.69; H 0 0 0.69',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g',
            rank=hvd.local_rank())

        # wave function
        self.wf = SlaterJastrow(self.mol, kinetic='jacobi',
                                configs='cas(2,2)',
                                cuda=False)

        # sampler
        self.sampler = Metropolis(
            nwalkers=200,
            nstep=200,
            step_size=0.2,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('atomic'),
            move={
                'type': 'all-elec',
                'proba': 'normal'})

        # optimizer
        self.opt = optim.Adam(self.wf.parameters(), lr=0.01)

        # solver
        self.solver = SolverMPI(wf=self.wf, sampler=self.sampler,
                                                 optimizer=self.opt, rank=hvd.rank())

        # ground state energy
        self.ground_state_energy = -1.16

        # ground state pos
        self.ground_state_pos = 0.69

    def test_single_point(self):
        self.solver.wf.ao.atom_coords[0, 2] = -self.ground_state_pos
        self.solver.wf.ao.atom_coords[1, 2] = self.ground_state_pos

        # sample and compute observables
        obs = self.solver.single_point()
        e, v = obs.energy, obs.variance

        e = e.data.item()
        v = v.data.item()

        assert np.isclose(e, -1.15, 0.2)
        assert 0 < v < 2

    def test_wf_opt(self):
        self.solver.wf.ao.atom_coords[0, 2] = -self.ground_state_pos
        self.solver.wf.ao.atom_coords[1, 2] = self.ground_state_pos

        self.solver.configure(track=['local_energy'], freeze=['ao', 'mo'],
                              loss='energy', grad='auto',
                              ortho_mo=False, clip_loss=False,
                              resampling={'mode': 'update',
                                          'resample_every': 1,
                                          'nstep_update': 50})
        self.solver.run(10)

        MPI.COMM_WORLD.barrier()

        self.solver.wf.load(self.solver.hdf5file, 'wf_opt')
        self.solver.wf.eval()

        obs = self.solver.single_point()

        e, v = obs.energy, obs.variance

        e = e.data.numpy()
        v = v.data.numpy()

        assert np.isclose(e, -1.15, 0.2)
        assert 0 < v < 2

    # def test_geo_opt(self):
    #     self.solver.wf.ao.atom_coords[0, 2].data = torch.as_tensor(-0.37)
    #     self.solver.wf.ao.atom_coords[1, 2].data = torch.as_tensor(0.37)
    #
    #     self.solver.configure(track=['local_energy'], freeze=['ao', 'mo'],
    #                           loss='energy', grad='auto',
    #                           ortho_mo=False, clip_loss=False,
    #                           resampling={'mode': 'update',
    #                                       'resample_every': 1,
    #                                       'nstep_update': 50})
    #     self.solver.geo_opt(5, nepoch_wf_init=10, nepoch_wf_update=5)
    #
    #     MPI.COMM_WORLD.barrier()
    #
    #     # load the best model
    #     self.solver.wf.load(self.solver.hdf5file, 'geo_opt')
    #     self.solver.wf.eval()
    #
    #     # sample and compute variables
    #     obs = self.solver.single_point()
    #     e, v = obs.energy, obs.variance
    #
    #     e = e.data.numpy()
    #     v = v.data.numpy()
    #
    #     assert np.isclose(e, -1.15, 0.2)
    #     assert (0 < v < 2.)


if __name__ == "__main__":
    t = TestH2Hvd()
    t.setUp()
    t.test_single_point()
    t.test_wf_opt()
    # t.test_geo_opt()



from tqdm import tqdm
import torch
from torch.autograd import Variable, grad
from torch.distributions import MultivariateNormal
import numpy as np

from .sampler_base import SamplerBase
from .. import log


class GeneralizedMetropolis(SamplerBase):

    def __init__(self, nwalkers=100, nstep=1000, step_size=3,
                 ntherm=-1, ndecor=1,
                 nelec=1, ndim=1,
                 init={'type': 'uniform', 'min': -5, 'max': 5},
                 cuda=False):
        """Generalized Metropolis Hasting sampler

        Args:
            nwalkers (int, optional): number of walkers. Defaults to 100.
            nstep (int, optional): number of steps. Defaults to 1000.
            step_size (int, optional): size of the steps. Defaults to 3.
            ntherm (int, optional): number of steps for thermalization. Defaults to -1.
            ndecor (int, optional): number of steps for decorelation. Defaults to 1.
            nelec (int, optional): number of electron. Defaults to 1.
            ndim (int, optional): number of dimensions. Defaults to 1.
            init (dict, optional): method to initialize the walkers. Defaults to {'type': 'uniform', 'min': -5, 'max': 5}.
            cuda (bool, optional): use cuda. Defaults to False.
        """

        SamplerBase.__init__(self, nwalkers, nstep,
                             step_size, ntherm, ndecor, nelec, ndim, init,
                             cuda)

    def __call__(self, pdf, pos=None, with_tqdm=True):
        """Generate a series of point using MC sampling

        Args:
            pdf (callable): probability distribution function to be sampled
            pos (torch.tensor, optional): position to start with.
                                          Defaults to None.
            with_tqdm (bool, optional): use tqdm to monitor progress

        Returns:
            torch.tensor: positions of the walkers
        """
        with torch.no_grad():

            if self.ntherm < 0:
                self.ntherm = self.nstep + self.ntherm

            self.walkers.initialize(pos=pos)

            xi = self.walkers.pos.clone()
            xi.requires_grad = True

            rhoi = pdf(xi)
            drifti = self.get_drift(pdf, xi)

            rhoi[rhoi == 0] = 1E-16
            pos, rate, idecor = [], 0, 0

            rng = tqdm(range(self.nstep),
                       desc='INFO:QMCTorch|  Sampling',
                       disable=not with_tqdm)

            for istep in rng:

                # new positions
                xf = self.move(drifti)

                # new function
                rhof = pdf(xf)
                driftf = self.get_drift(pdf, xf)
                rhof[rhof == 0.] = 1E-16

                # transtions
                Tif = self.trans(xi, xf, driftf)
                Tfi = self.trans(xf, xi, drifti)
                pmat = (Tif * rhof) / (Tfi * rhoi).double()

                # accept the moves
                index = self._accept(pmat)

                # acceptance rate
                rate += index.byte().sum().float() / self.nwalkers

                # update position/function value
                xi[index, :] = xf[index, :]
                rhoi[index] = rhof[index]
                rhoi[rhoi == 0] = 1E-16

                drifti[index, :] = driftf[index, :]

                if (istep >= self.ntherm):
                    if (idecor % self.ndecor == 0):
                        pos.append(xi.clone().detach())
                    idecor += 1

            log.options(style='percent').debug("  Acceptance rate %1.3f" %
                                               (rate / self.nstep * 100))

            self.walkers.pos.data = xi.data

        return torch.cat(pos).requires_grad_()

    def move(self, drift):
        """Move electron one at a time in a vectorized way.

        Args:
            drift (torch.tensor): drift velocity of the walkers

        Returns:
            torch.tensor: new positions of the walkers
        """

        # clone and reshape data : Nwlaker, Nelec, Ndim
        new_pos = self.walkers.pos.clone()
        new_pos = new_pos.view(self.nwalkers,
                               self.nelec, self.ndim)

        # get indexes
        index = torch.LongTensor(self.nwalkers).random_(
            0, self.nelec)

        new_pos[range(self.nwalkers), index,
                :] += self._move(drift, index)

        return new_pos.view(self.nwalkers, self.nelec * self.ndim)

    def _move(self, drift, index):
        """Move a walker.

        Args:
            drift (torch.tensor): drift velocity
            index (int): indx of the electron to move

        Returns:
            torch.tensor: position of the walkers
        """

        d = drift.view(self.nwalkers,
                       self.nelec, self.ndim)

        mv = MultivariateNormal(torch.zeros(self.ndim), np.sqrt(
            self.step_size) * torch.eye(self.ndim))

        return self.step_size * d[range(self.nwalkers), index, :] \
            + mv.sample((self.nwalkers, 1)).squeeze()

    def trans(self, xf, xi, drifti):
        """transform the positions

        Args:
            xf ([type]): [description]
            xi ([type]): [description]
            drifti ([type]): [description]

        Returns:
            [type]: [description]
        """
        a = (xf - xi - drifti * self.step_size).norm(dim=1)
        return torch.exp(- 0.5 * a / self.step_size)

    def get_drift(self, pdf, x):
        """Compute the drift velocity

        Args:
            pdf (callable): function that returns the density
            x (torch.tensor): positions of the walkers

        Returns:
            torch.tensor: drift velocity
        """
        with torch.enable_grad():

            x.requires_grad = True
            rho = pdf(x).view(-1, 1)
            z = Variable(torch.ones_like(rho))
            grad_rho = grad(rho, x,
                            grad_outputs=z,
                            only_inputs=True)[0]
            return 0.5 * grad_rho / rho

    def _accept(self, P):
        """accept the move or not

        Args:
            P (torch.tensor): probability of each move

        Returns:
            torch.tensor: the indx of the accepted moves
        """
        P[P > 1] = 1.0
        tau = torch.rand(self.walkers.nwalkers).double()
        index = (P - tau >= 0).reshape(-1)
        return index.type(torch.bool)



from tests.wavefunction.test_slaterjastrow import TestSlaterJastrow
import unittest

import numpy as np
import torch
import torch.optim as optim

from qmctorch.sampler import Metropolis
from qmctorch.solver import Solver
from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrowBackFlow
from qmctorch.utils import set_torch_double_precision


class TestLiHBackFlowPySCF(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(0)
        np.random.seed(0)
        set_torch_double_precision()

        # molecule
        self.mol = Molecule(
            atom='Li 0 0 0; H 0 0 3.015',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g')

        # wave function
        self.wf = SlaterJastrowBackFlow(self.mol, kinetic='jacobi',
                                        configs='single_double(2,2)',
                                        orbital_dependent_backflow=True,
                                        include_all_mo=True)

        # fc weights
        self.wf.fc.weight.data = torch.rand(self.wf.fc.weight.shape)

        # jastrow weights
        self.wf.jastrow.jastrow_kernel.weight.data = torch.rand(
            self.wf.jastrow.jastrow_kernel.weight.shape)

        # sampler
        self.sampler = Metropolis(
            nwalkers=500,
            nstep=200,
            step_size=0.05,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('normal'),
            move={
                'type': 'all-elec',
                'proba': 'normal'})

        # optimizer
        self.opt = optim.Adam(self.wf.parameters(), lr=0.01)

        # solver
        self.solver = Solver(wf=self.wf, sampler=self.sampler,
                                          optimizer=self.opt)

        # artificial pos
        self.nbatch = 10
        self.pos = torch.as_tensor(np.random.rand(
            self.nbatch, self.wf.nelec*3))
        self.pos.requires_grad = True

    def test_0_wavefunction(self):

        eauto = self.wf.kinetic_energy_autograd(self.pos)
        ejac = self.wf.kinetic_energy_jacobi(self.pos)
        print(torch.stack([eauto, ejac], axis=1).squeeze())
        assert torch.allclose(
            eauto.data, ejac.data, rtol=1E-4, atol=1E-4)

    def test1_single_point(self):

        # sample and compute observables
        obs = self.solver.single_point()
        e, v = obs.energy, obs.variance

    def test2_wf_opt_grad_auto(self):
        self.solver.sampler = self.sampler

        self.solver.configure(track=['local_energy'],
                              loss='energy', grad='auto')
        obs = self.solver.run(5)

    def test3_wf_opt_grad_manual(self):
        self.solver.sampler = self.sampler

        self.solver.configure(track=['local_energy', 'parameters'],
                              loss='energy', grad='manual')
        obs = self.solver.run(5)


if __name__ == "__main__":
    # unittest.main()
    t = TestLiHBackFlowPySCF()
    t.setUp()
    t.test3_wf_opt_grad_manual()



import torch
from torch import nn
from .backflow_kernel_base import BackFlowKernelBase


class BackFlowKernelAutoInverse(BackFlowKernelBase):

    def __init__(self, mol, cuda, order=2):
        """Compute the back flow kernel, i.e. the function
        f(rij) where rij is the distance between electron i and j
        This kernel is used in the backflow transformation
        .. math:
            q_i = r_i + \\sum_{j\\neq i} f(r_{ij}) (r_i-r_j)
        """
        super().__init__(mol, cuda)
        self.order = order
        self.fc = nn.Linear(order, 1, bias=False)
        self.fc.weight.data *= 0.
        self.fc.weight.data[0, 0] = 1.

        self.weight = nn.Parameter(
            torch.as_tensor([1E-3]))

    def _backflow_kernel(self, ree):
        """Computes the kernel via autodiff

        Args:
            ree ([type]): [description]

        Returns:
            [type]: [description]
        """
        eye = torch.eye(self.nelec, self.nelec).to(self.device)
        mask = torch.ones_like(ree) - eye
        return self.weight * mask * (1./(ree+eye) - eye)



import unittest

import numpy as np
import torch

from qmctorch.utils import set_torch_double_precision
from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow


class TestGTO2STOFit(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(101)
        np.random.seed(101)

        set_torch_double_precision()

        # molecule
        mol = Molecule(
            atom='C 0 0 0',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g',
            redo_scf=True)

        self.wf = SlaterJastrow(mol, kinetic='auto',
                                configs='ground_state').gto2sto()

        self.pos = -0.25 + 0.5 * \
            torch.as_tensor(np.random.rand(10, 18))
        self.pos.requires_grad = True

    def test_forward(self):

        wfvals = self.wf(self.pos)
        ref = torch.as_tensor([[-8.4430e-06],
                               [1.5092e-02],
                               [3.3809e-03],
                               [9.7981e-03],
                               [-6.8513e-02],
                               [-4.6836e-03],
                               [-3.2847e-04],
                               [2.3636e-02],
                               [5.5934e-04],
                               [1.3205e-02]])
        assert torch.allclose(wfvals.data, ref, rtol=1E-4, atol=1E-4)


if __name__ == "__main__":
    unittest.main()



from time import time
from types import SimpleNamespace

import torch
from torch.utils.data import DataLoader

from qmctorch.utils import (DataSet, Loss, OrthoReg, add_group_attr,
                            dump_to_hdf5)

from .. import log
from .solver import Solver

try:
    import horovod.torch as hvd
except ModuleNotFoundError:
    pass


def logd(rank, *args):
    if rank == 0:
        log.info(*args)


class SolverMPI(Solver):

    def __init__(self, wf=None, sampler=None, optimizer=None,
                 scheduler=None, output=None, rank=0):
        """Distributed QMC solver

        Args:
            wf (qmctorch.WaveFunction, optional): wave function. Defaults to None.
            sampler (qmctorch.sampler, optional): Sampler. Defaults to None.
            optimizer (torch.optim, optional): optimizer. Defaults to None.
            scheduler (torch.optim, optional): scheduler. Defaults to None.
            output (str, optional): hdf5 filename. Defaults to None.
            rank (int, optional): rank of he process. Defaults to 0.
        """

        super().__init__(wf, sampler,
                         optimizer, scheduler, output, rank)

        hvd.broadcast_optimizer_state(self.opt, root_rank=0)
        self.opt = hvd.DistributedOptimizer(
            self.opt, named_parameters=self.wf.named_parameters())

        self.sampler.nwalkers //= hvd.size()
        self.sampler.walkers.nwalkers //= hvd.size()

    def run(self, nepoch, batchsize=None, loss='energy',
            clip_loss=False, grad='manual', hdf5_group='wf_opt',
            num_threads=1, chkpt_every=None):
        """Run the optimization

        Args:
            nepoch (int): Number of optimization step
            batchsize (int, optional): Number of sample in a mini batch.
                                       If None, all samples are used.
                                       Defaults to None.
            loss (str, optional): method to compute the loss: variance or energy.
                                  Defaults to 'energy'.
            clip_loss (bool, optional): Clip the loss values at +/- 5std.
                                        Defaults to False.
            grad (str, optional): method to compute the gradients: 'auto' or 'manual'.
                                  Defaults to 'auto'.
            hdf5_group (str, optional): name of the hdf5 group where to store the data.
                                    Defaults to 'wf_opt'
        """

        logd(hvd.rank(), '')
        logd(hvd.rank(),
             '  Distributed Optimization on {num} process'.format(num=hvd.size()))
        log.info('   - Process {id} using {nw} walkers'.format(
                 id=hvd.rank(), nw=self.sampler.nwalkers))

        # observable
        if not hasattr(self, 'observable'):
            self.track_observable(['local_energy'])

        self.evaluate_gradient = {
            'auto': self.evaluate_grad_auto,
            'manual': self.evaluate_grad_manual}[grad]

        if 'lpos_needed' not in self.opt.__dict__.keys():
            self.opt.lpos_needed = False

        self.wf.train()

        hvd.broadcast_parameters(self.wf.state_dict(), root_rank=0)
        torch.set_num_threads(num_threads)

        # get the loss
        self.loss = Loss(self.wf, method=loss, clip=clip_loss)
        self.loss.use_weight = (
            self.resampling_options.resample_every > 1)

        # orthogonalization penalty for the MO coeffs
        self.ortho_loss = OrthoReg()

        self.prepare_optimization(batchsize, chkpt_every)
        # log data
        if hvd.rank() == 0:
            self.log_data_opt(nepoch, 'wave function optimization')

        # sample the wave function
        if hvd.rank() == 0:
            pos = self.sampler(self.wf.pdf)
        else:
            pos = self.sampler(self.wf.pdf, with_tqdm=False)

        # requried to build the distributed data container
        pos.requires_grad_(False)

        # handle the batch size
        if batchsize is None:
            batchsize = len(pos)

        # get the initial observable
        if hvd.rank() == 0:
            self.store_observable(pos)

        # change the number of steps/walker size
        _nstep_save = self.sampler.nstep
        _ntherm_save = self.sampler.ntherm
        _nwalker_save = self.sampler.walkers.nwalkers
        if self.resampling_options.mode == 'update':
            self.sampler.ntherm = -1
            self.sampler.nstep = self.resampling_options.nstep_update
            self.sampler.walkers.nwalkers = pos.shape[0]
            self.sampler.nwalkers = pos.shape[0]

        # create the data loader
        self.dataset = DataSet(pos)

        if self.cuda:
            kwargs = {'num_workers': num_threads, 'pin_memory': True}
        else:
            kwargs = {'num_workers': num_threads}

        self.dataloader = DataLoader(self.dataset,
                                     batch_size=batchsize,
                                     **kwargs)
        min_loss = 1E3

        for n in range(nepoch):

            tstart = time()
            logd(hvd.rank(), '')
            logd(hvd.rank(), '  epoch %d' % n)

            cumulative_loss = 0.

            for ibatch, data in enumerate(self.dataloader):

                # get data
                lpos = data.to(self.device)
                lpos.requires_grad = True

                # get the gradient
                loss, eloc = self.evaluate_gradient(lpos)
                cumulative_loss += loss

                # optimize the parameters
                self.optimization_step(lpos)

                # observable
                if hvd.rank() == 0:
                    self.store_observable(
                        pos, local_energy=eloc, ibatch=ibatch)

            cumulative_loss = self.metric_average(cumulative_loss,
                                                  'cum_loss')

            if hvd.rank() == 0:
                if n == 0 or cumulative_loss < min_loss:
                    self.observable.models.best = dict(
                        self.wf.state_dict())
                min_loss = cumulative_loss

                if self.chkpt_every is not None:
                    if (n > 0) and (n % chkpt_every == 0):
                        self.save_checkpoint(n, cumulative_loss)

                self.print_observable(cumulative_loss)

            # resample the data
            pos = self.resample(n, pos)
            pos.requires_grad = False

            # scheduler step
            if self.scheduler is not None:
                self.scheduler.step()

            logd(hvd.rank(), '  epoch done in %1.2f sec.' %
                 (time()-tstart))

        # restore the sampler number of step
        self.sampler.nstep = _nstep_save
        self.sampler.ntherm = _ntherm_save
        self.sampler.walkers.nwalkers = _nwalker_save
        self.sampler.nwalkers = _nwalker_save

        if hvd.rank() == 0:
            dump_to_hdf5(self.observable, self.hdf5file, hdf5_group)
            add_group_attr(self.hdf5file, hdf5_group, {'type': 'opt'})

        return self.observable

    def single_point(self, with_tqdm=True, hdf5_group='single_point'):
        """Performs a single point calculation

        Args:
            with_tqdm (bool, optional): use tqdm for samplig. Defaults to True.
            hdf5_group (str, optional): hdf5 group where to store the data.
                                        Defaults to 'single_point'.

        Returns:
            SimpleNamespace: contains the local energy, positions, ...
        """

        logd(hvd.rank(), '')
        logd(hvd.rank(), '  Single Point Calculation : {nw} walkers | {ns} steps'.format(
             nw=self.sampler.nwalkers, ns=self.sampler.nstep))

        # check if we have to compute and store the grads
        grad_mode = torch.no_grad()
        if self.wf.kinetic == 'auto':
            grad_mode = torch.enable_grad()

        # distribute the calculation
        num_threads = 1
        hvd.broadcast_parameters(self.wf.state_dict(), root_rank=0)
        torch.set_num_threads(num_threads)

        with grad_mode:

            # sample the wave function
            pos = self.sampler(self.wf.pdf)
            if self.wf.cuda and pos.device.type == 'cpu':
                pos = pos.to(self.device)

            # compute energy/variance/error
            eloc = self.wf.local_energy(pos)
            e, s, err = torch.mean(eloc), torch.var(
                eloc), self.wf.sampling_error(eloc)

            # gather all data
            eloc_all = hvd.allgather(eloc, name='local_energies')
            e, s, err = torch.mean(eloc_all), torch.var(
                eloc_all), self.wf.sampling_error(eloc_all)

            # print
            if hvd.rank() == 0:
                log.options(style='percent').info(
                    '  Energy   : %f +/- %f' % (e.detach().item(), err.detach().item()))
                log.options(style='percent').info(
                    '  Variance : %f' % s.detach().item())

            # dump data to hdf5
            obs = SimpleNamespace(
                pos=pos,
                local_energy=eloc_all,
                energy=e,
                variance=s,
                error=err
            )

            # dump to file
            if hvd.rank() == 0:

                dump_to_hdf5(obs,
                             self.hdf5file,
                             root_name=hdf5_group)
                add_group_attr(self.hdf5file, hdf5_group,
                               {'type': 'single_point'})

        return obs

    @staticmethod
    def metric_average(val, name):
        """Average a give quantity over all processes

        Arguments:
            val {torch.tensor} -- data to average
            name {str} -- name of the data

        Returns:
            torch.tensor -- Averaged quantity
        """
        tensor = val.clone().detach()
        avg_tensor = hvd.allreduce(tensor, name=name)
        return avg_tensor.item()




from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrowBackFlow

from qmctorch.wavefunction.orbitals.backflow.kernels import BackFlowKernelInverse
from qmctorch.wavefunction.jastrows.elec_elec.kernels import PadeJastrowKernel

from qmctorch.utils import set_torch_double_precision

from torch.autograd import grad, Variable

import numpy as np
import torch
import unittest

torch.set_default_tensor_type(torch.DoubleTensor)


def hess(out, pos):
    # compute the jacobian
    z = Variable(torch.ones(out.shape))
    jacob = grad(out, pos,
                 grad_outputs=z,
                 only_inputs=True,
                 create_graph=True)[0]

    # compute the diagonal element of the Hessian
    z = Variable(torch.ones(jacob.shape[0]))
    hess = torch.zeros(jacob.shape)

    for idim in range(jacob.shape[1]):

        tmp = grad(jacob[:, idim], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, idim] = tmp[:, idim]

    return hess


class TestSlaterJastrowOrbitalDependentBackFlow(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(101)
        np.random.seed(101)

        set_torch_double_precision()

        # molecule
        mol = Molecule(
            atom='Li 0 0 0; H 0 0 3.015',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g',
            redo_scf=True)

        self.wf = SlaterJastrowBackFlow(mol,
                                        kinetic='jacobi',
                                        jastrow_kernel=PadeJastrowKernel,
                                        include_all_mo=True,
                                        configs='single_double(2,2)',
                                        backflow_kernel=BackFlowKernelInverse,
                                        orbital_dependent_backflow=True
                                        )

        # change the weights
        for ker in self.wf.ao.backflow_trans.backflow_kernel.orbital_dependent_kernel:
            ker.weight.data[0] = torch.rand(1)

        self.random_fc_weight = torch.rand(self.wf.fc.weight.shape)
        self.wf.fc.weight.data = self.random_fc_weight

        self.nbatch = 5
        self.pos = torch.Tensor(np.random.rand(
            self.nbatch,  self.wf.nelec*3))
        self.pos.requires_grad = True

    def test_forward(self):
        wfvals = self.wf(self.pos)

    def test_antisymmetry(self):
        """Test that the wf values are antisymmetric
        wrt exchange of 2 electrons of same spin."""
        wfvals_ref = self.wf(self.pos)

        if self.wf.nelec < 4:
            print(
                'Warning : antisymmetry cannot be tested with \
                    only %d electrons' % self.wf.nelec)
            return

        # test spin up
        pos_xup = self.pos.clone()
        perm_up = list(range(self.wf.nelec))
        perm_up[0] = 1
        perm_up[1] = 0
        pos_xup = pos_xup.reshape(self.nbatch, self.wf.nelec, 3)
        pos_xup = pos_xup[:, perm_up, :].reshape(
            self.nbatch, self.wf.nelec*3)

        wfvals_xup = self.wf(pos_xup)
        assert(torch.allclose(wfvals_ref, -1*wfvals_xup))

    def test_jacobian_mo(self):
        """Jacobian of the BF MOs."""

        mo = self.wf.pos2mo(self.pos)
        dmo = self.wf.pos2mo(self.pos, derivative=1)

        dmo_grad = grad(
            mo, self.pos, grad_outputs=torch.ones_like(mo))[0]
        assert(torch.allclose(dmo.sum(), dmo_grad.sum()))

        psum_mo = dmo.sum(-1).sum(-1)
        psum_mo_grad = dmo_grad.view(
            self.nbatch, self.wf.nelec, 3).sum(-1)
        psum_mo_grad = psum_mo_grad.T
        assert(torch.allclose(psum_mo, psum_mo_grad))

    def test_grad_mo(self):
        """Gradients of the BF MOs."""

        mo = self.wf.pos2mo(self.pos)

        dao = self.wf.ao(self.pos, derivative=1, sum_grad=False)
        dmo = self.wf.ao2mo(dao)

        dmo_grad = grad(
            mo, self.pos,
            grad_outputs=torch.ones_like(mo))[0]
        assert(torch.allclose(dmo.sum(), dmo_grad.sum()))

        dmo = dmo.sum(-1).sum(-1)
        dmo_grad = dmo_grad.T

        assert(torch.allclose(dmo, dmo_grad))

    def test_hess_mo(self):
        """Hessian of the MOs."""
        val = self.wf.pos2mo(self.pos)

        d2val_grad = hess(val, self.pos)
        d2ao = self.wf.ao(self.pos, derivative=2, sum_hess=False)
        d2val = self.wf.ao2mo(d2ao)

        assert(torch.allclose(d2val.sum(), d2val_grad.sum()))

        d2val = d2val.reshape(4, 3, 5, 4, 6).sum(1).sum(-1).sum(-1)
        d2val_grad = d2val_grad.view(
            self.nbatch, self.wf.nelec, 3).sum(-1)
        d2val_grad = d2val_grad.T
        assert(torch.allclose(d2val, d2val_grad))

    def test_grad_wf(self):
        pass

        # grad_auto = self.wf.gradients_autograd(self.pos)
        # grad_jac = self.wf.gradients_jacobi(self.pos)

        # assert torch.allclose(
        #     grad_auto.data, grad_jac.data, rtol=1E-4, atol=1E-4)

    def test_local_energy(self):

        self.wf.kinetic_energy = self.wf.kinetic_energy_autograd
        eloc_auto = self.wf.local_energy(self.pos)

        self.wf.kinetic_energy = self.wf.kinetic_energy_jacobi
        eloc_jac = self.wf.local_energy(self.pos)

        assert torch.allclose(
            eloc_auto.data, eloc_jac.data, rtol=1E-4, atol=1E-4)

    def test_kinetic_energy(self):

        eauto = self.wf.kinetic_energy_autograd(self.pos)
        ejac = self.wf.kinetic_energy_jacobi(self.pos)

        print(ejac)
        print(eauto)

        assert torch.allclose(
            eauto.data, ejac.data, rtol=1E-4, atol=1E-4)


if __name__ == "__main__":
    # t = TestSlaterJastrowOrbitalDependentBackFlow()
    # t.setUp()
    # t.test_antisymmetry()
    # t.test_hess_mo()
    # t.test_grad_mo()
    # t.test_kinetic_energy()
    unittest.main()



import torch


def get_scaled_distance(kappa, r):
    """compute the scaled distance

    .. math::
        u_{ij} = \\frac{1-e^{-\\kappa r_{ij}}}{\\kappa}

    Args:
        kappa (float) : scaling factor
        r (torch.tensor): matrix of the e-e distances
                            Nbatch x Nelec x Nelec

    Returns:
        torch.tensor: values of the scaled distance
                        Nbatch, Nelec, Nelec
    """
    return (1. - torch.exp(-kappa * r))/kappa


def get_der_scaled_distance(kappa, r, dr):
    """Returns the derivative of the scaled distances

    .. math::
        \\frac{d u}{d x_i} = \\frac{d r_{ij}}{d x_i} e^{-\\kappa r_{ij}}
    Args:
        kappa (float) : scaling factor
        r (torch.tensor): matrix of the e-e distances
                            Nbatch x Nelec x Nelec

        dr (torch.tensor): matrix of the derivative of the e-e distances
                            Nbatch x Ndim x Nelec x Nelec

    Returns:
        torch.tensor : deriative of the scaled distance
                        Nbatch x Ndim x Nelec x Nelec
    """
    return dr * torch.exp(-kappa * r.unsqueeze(1))


def get_second_der_scaled_distance(kappa, r, dr, d2r):
    """computes the second derivative of the scaled distances

    .. math::
        \\frac{d^2u_{ij}}{d x_i^2} = \\frac{d^2r_{ij}}{d x_i^2} -\\kappa \\left( \\frac{d r_{ij}}{d x_i} \\right)^2 e^{-\\kappa r_{ij}}

    Args:
        kappa (float) : scaling factor
        r (torch.tensor): unsqueezed matrix of the e-e distances
                            Nbatch x Nelec x Nelec
        dr (torch.tensor): matrix of the derivative of the e-e distances
                            Nbatch x Ndim x Nelec x Nelec
        d2r (torch.tensor): matrix of the 2nd derivative of
                            the e-e distances
                            Nbatch x Ndim x Nelec x Nelec

    Returns:
        torch.tensor : second deriative of the scaled distance
                        Nbatch x Ndim x Nelec x Nelec
    """
    return (d2r - kappa * dr * dr) * torch.exp(-kappa*r.unsqueeze(1))



import h5py
import torch
from torch.autograd import Variable, grad


class WaveFunction(torch.nn.Module):

    def __init__(self, nelec, ndim, kinetic='auto', cuda=False):

        super(WaveFunction, self).__init__()

        self.ndim = ndim
        self.nelec = nelec
        self.ndim_tot = self.nelec * self.ndim
        self.kinetic = kinetic
        self.cuda = cuda
        self.device = torch.device('cpu')
        if self.cuda:
            self.device = torch.device('cuda')
        self.kinetic_energy = self.kinetic_energy_autograd
        self.gradients = self.gradients_autograd

    def forward(self, x):
        ''' Compute the value of the wave function.
        for a multiple conformation of the electrons

        Args:
            parameters : variational param of the wf
            pos: position of the electrons

        Returns: values of psi
        '''

        raise NotImplementedError()

    def electronic_potential(self, pos):
        r"""Computes the electron-electron term

        .. math:
            V_{ee} = \sum_{e_1} \sum_{e_2} \\frac{1}{r_{e_1e_2}}

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)

        Returns:
            torch.tensor: values of the electon-electron energy at each sampling points
        """

        pot = torch.zeros(pos.shape[0], device=self.device)

        for ielec1 in range(self.nelec - 1):
            epos1 = pos[:, ielec1 *
                        self.ndim:(ielec1 + 1) * self.ndim]
            for ielec2 in range(ielec1 + 1, self.nelec):
                epos2 = pos[:, ielec2 *
                            self.ndim:(ielec2 + 1) * self.ndim]
                r = torch.sqrt(((epos1 - epos2)**2).sum(1))  # + 1E-12
                pot += (1. / r)
        return pot.view(-1, 1)

    def nuclear_potential(self, pos):
        r"""Computes the electron-nuclear term

        .. math:
            V_{en} = - \sum_e \sum_n \\frac{Z_n}{r_{en}}

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)

        Returns:
            torch.tensor: values of the electon-nuclear energy at each sampling points
        """

        p = torch.zeros(pos.shape[0], device=self.device)
        for ielec in range(self.nelec):
            istart = ielec * self.ndim
            iend = (ielec + 1) * self.ndim
            pelec = pos[:, istart:iend]
            for iatom in range(self.natom):
                patom = self.ao.atom_coords[iatom, :]
                Z = self.ao.atomic_number[iatom]
                r = torch.sqrt(((pelec - patom)**2).sum(1))  # + 1E-12
                p += -Z / r
        return p.view(-1, 1)

    def nuclear_repulsion(self):
        r"""Computes the nuclear-nuclear repulsion term

        .. math:
            V_{nn} = \sum_{n_1} \sum_{n_2} \\frac{Z_1Z_2}{r_{n_1n_2}}

        Returns:
            torch.tensor: values of the nuclear-nuclear energy at each sampling points
        """

        vnn = 0.
        for at1 in range(self.natom - 1):
            c0 = self.ao.atom_coords[at1, :]
            Z0 = self.ao.atomic_number[at1]
            for at2 in range(at1 + 1, self.natom):
                c1 = self.ao.atom_coords[at2, :]
                Z1 = self.ao.atomic_number[at2]
                rnn = torch.sqrt(((c0 - c1)**2).sum())
                vnn += Z0 * Z1 / rnn
        return vnn

    def gradients_autograd(self, pos, pdf=False):
        """Computes the gradients of the wavefunction (or density)
        w.r.t the values of the pos.

        Args:
            pos (torch.tensor): positions of the walkers
            pdf (bool, optional) : if true compute the grads of the density

        Returns:
            torch.tensor: values of the gradients
        """
        out = self.forward(pos)

        # compute the grads
        grads = grad(out, pos,
                     grad_outputs=torch.ones_like(out),
                     only_inputs=True)[0]

        # if we return grad of pdf
        if pdf:
            grads = 2*grads*out

        return grads

    def kinetic_energy_autograd(self, pos):
        """Compute the kinetic energy through the 2nd derivative
        w.r.t the value of the pos.

        Args:
            pos (torch.tensor): positions of the walkers

        Returns:
            values of nabla^2 * Psi
        """

        out = self.forward(pos)

        # compute the jacobian
        z = torch.ones_like(out)
        jacob = grad(out, pos,
                     grad_outputs=z,
                     only_inputs=True,
                     create_graph=True)[0]

        # compute the diagonal element of the Hessian
        z = Variable(torch.ones(jacob.shape[0])).to(self.device)
        hess = torch.zeros(jacob.shape[0]).to(self.device)

        for idim in range(jacob.shape[1]):

            tmp = grad(jacob[:, idim], pos,
                       grad_outputs=z,
                       only_inputs=True,
                       create_graph=True)[0]

            hess += tmp[:, idim]

        return -0.5 * hess.view(-1, 1) / out

    def local_energy(self, pos):
        """Computes the local energy

         .. math::
             E = K(R) + V_{ee}(R) + V_{en}(R) + V_{nn}

         Args:
             pos (torch.tensor): sampling points (Nbatch, 3*Nelec)

         Returns:
             [torch.tensor]: values of the local enrgies at each sampling points

         Examples::
             >>> mol = Molecule('h2.xyz', calculator='adf', basis = 'dzp')
             >>> wf = SlaterJastrow(mol, configs='cas(2,2)')
             >>> pos = torch.rand(500,6)
             >>> vals = wf.local_energy(pos)

         Note:
            by default kinetic_energy refers to kinetic_energy_autograd
            users can overwrite it to poit to any other methods
            see kinetic_energy_jacobi in wf_orbital
         """

        ke = self.kinetic_energy(pos)

        return ke  \
            + self.nuclear_potential(pos)  \
            + self.electronic_potential(pos) \
            + self.nuclear_repulsion()

    def energy(self, pos):
        '''Total energy for the sampling points.'''
        return torch.mean(self.local_energy(pos))

    def variance(self, pos):
        '''Variance of the energy at the sampling points.'''
        return torch.var(self.local_energy(pos))

    def sampling_error(self, eloc):
        '''Compute the statistical uncertainty.
        Assuming the samples are uncorrelated.'''
        Npts = eloc.shape[0]
        return torch.sqrt(eloc.var() / Npts)

    def _energy_variance(self, pos):
        '''Return energy and variance.'''
        el = self.local_energy(pos)
        return torch.mean(el), torch.var(el)

    def _energy_variance_error(self, pos):
        '''Return energy variance and sampling error.'''
        el = self.local_energy(pos)
        return torch.mean(el), torch.var(el), self.sampling_error(el)

    def pdf(self, pos, return_grad=False):
        '''density of the wave function.'''
        if return_grad:
            return self.gradients(pos, pdf=True)
        else:
            return (self.forward(pos)**2).reshape(-1)

    def get_number_parameters(self):
        """Computes the total number of parameters."""
        nparam = 0
        for name, param in self.named_parameters():
            if param.requires_grad:
                nparam += param.data.numel()
        return nparam

    def load(self, filename, group='wf_opt', model='best'):
        """Load trained parameters

        Args:
            filename (str): hdf5 filename
            group (str, optional): group in the hdf5 file where the model is stored.
                                   Defaults to 'wf_opt'.
            model (str, optional): 'best' or ' last'. Defaults to 'best'.
        """
        f5 = h5py.File(filename, 'r')
        grp = f5[group]['models'][model]
        data = dict()
        for name, val in grp.items():
            data[name] = torch.as_tensor(val)
        self.load_state_dict(data)
        f5.close()



import torch
from torch import nn
import numpy as np
from .jastrow_kernel_electron_electron_base import JastrowKernelElectronElectronBase


class FullyConnectedJastrowKernel(JastrowKernelElectronElectronBase):

    def __init__(self,  nup, ndown, cuda,
                 size1=16, size2=8,
                 activation=torch.nn.Sigmoid(),
                 include_cusp_weight=True):
        """Defines a fully connected jastrow factors."""

        super().__init__(nup, ndown, cuda)

        self.cusp_weights = None

        self.fc1 = nn.Linear(1, size1, bias=False)
        self.fc2 = nn.Linear(size1, size2, bias=False)
        self.fc3 = nn.Linear(size2, 1, bias=False)

        eps = 1E-6
        self.fc1.weight.data *= eps
        self.fc2.weight.data *= eps
        self.fc3.weight.data *= eps

        self.nl_func = activation
        #self.nl_func = lambda x:  x

        self.prefac = torch.rand(1)

        self.cusp_weights = self.get_static_weight()
        self.requires_autograd = True
        self.get_var_weight()

        self.include_cusp_weight = include_cusp_weight

    def get_var_weight(self):
        """define the variational weight."""

        nelec = self.nup + self.ndown

        self.var_cusp_weight = nn.Parameter(
            torch.as_tensor([0., 0.]))

        idx_pair = []
        for i in range(nelec-1):
            ispin = 0 if i < self.nup else 1
            for j in range(i+1, nelec):
                jspin = 0 if j < self.nup else 1

                if ispin == jspin:
                    idx_pair.append(0)
                else:
                    idx_pair.append(1)
        self.idx_pair = torch.as_tensor(idx_pair).to(self.device)

    def get_static_weight(self):
        """Get the matrix of static weights

        Returns:
            torch.tensor: static weight (0.5 (0.25) for parallel(anti) spins
        """

        bup = torch.cat((0.25 * torch.ones(self.nup, self.nup), 0.5 *
                         torch.ones(self.nup, self.ndown)), dim=1)

        bdown = torch.cat((0.5 * torch.ones(self.ndown, self.nup), 0.25 *
                           torch.ones(self.ndown, self.ndown)), dim=1)

        static_weight = torch.cat((bup, bdown), dim=0).to(self.device)

        mask_tri_up = torch.triu(torch.ones_like(
            static_weight), diagonal=1).type(torch.BoolTensor).to(self.device)
        static_weight = static_weight.masked_select(mask_tri_up)

        return static_weight

    def forward(self, x):
        """Compute the kernel values

        Args:
            x (torch.tensor): e-e distance Nbatch, Nele_pairs

        Returns:
            torch.tensor: values of the f_ij
        """
        nbatch, npairs = x.shape
        # w = (x*self.cusp_weights)

        # reshape the input so that all elements are considered
        # independently of each other
        x = x.reshape(-1, 1)

        x = self.fc1(x)
        x = self.nl_func(x)
        x = self.fc2(x)
        x = self.nl_func(x)
        x = self.fc3(x)
        x = self.nl_func(x)

        # reshape to the original shape
        x = x.reshape(nbatch, npairs)

        # add the cusp weight
        if self.include_cusp_weight:
            x = x + self.var_cusp_weight[self.idx_pair]
        return x



__all__ = [
    'SamplerBase',
    'Metropolis',
    'Hamiltonian',
    'GeneralizedMetropolis']

from .sampler_base import SamplerBase
from .metropolis import Metropolis
from .hamiltonian import Hamiltonian
from .generalized_metropolis import GeneralizedMetropolis



import torch
from torch import nn
from torch.autograd import grad, Variable


class BackFlowKernelBase(nn.Module):

    def __init__(self, mol, cuda):
        """Compute the back flow kernel, i.e. the function
        f(rij) where rij is the distance between electron i and j
        This kernel is used in the backflow transformation
        .. math:
            q_i = r_i + \\sum_{j\\neq i} f(r_{ij}) (r_i-r_j)
        """
        super().__init__()
        self.nelec = mol.nelec
        self.cuda = cuda
        self.device = torch.device('cpu')
        if self.cuda:
            self.device = torch.device('cuda')

    def forward(self, ree, derivative=0):
        """Computes the desired values of the kernel
         Args:
            ree (torch.tensor): e-e distance Nbatch x Nelec x Nelec
            derivative (int): derivative requried 0, 1, 2

        Returns:
            torch.tensor : f(r) Nbatch x Nelec x Nelec
        """

        if derivative == 0:
            return self._backflow_kernel(ree)

        elif derivative == 1:
            return self._backflow_kernel_derivative(ree)

        elif derivative == 2:
            return self._backflow_kernel_second_derivative(ree)

        else:
            raise ValueError(
                'derivative of the kernel must be 0, 1 or 2')

    def _backflow_kernel(self, ree):
        """Computes the kernel via autodiff

        Args:
            ree ([type]): [description]

        Returns:
            [type]: [description]
        """
        raise NotImplementedError(
            'Please implement the backflow kernel')

    def _backflow_kernel_derivative(self, ree):
        """Computes the first derivative of the kernel via autodiff

        Args:
            ree ([type]): [description]

        Returns:
            [type]: [description]
        """
        if ree.requires_grad == False:
            ree.requires_grad = True

        with torch.enable_grad():
            kernel_val = self._backflow_kernel(ree)

        return self._grad(kernel_val, ree)

    def _backflow_kernel_second_derivative(self, ree):
        """Computes the second derivative of the kernel via autodiff

        Args:
            ree ([type]): [description]

        Returns:
            [type]: [description]
        """
        if ree.requires_grad == False:
            ree.requires_grad = True

        with torch.enable_grad():

            kernel_val = self._backflow_kernel(ree)
            hess_val, _ = self._hess(kernel_val, ree)

        return hess_val

    @staticmethod
    def _grad(val, ree):
        """Get the gradients of the kernel.

        Args:
            ree ([type]): [description]

        Returns:
            [type]: [description]
        """
        return grad(val, ree, grad_outputs=torch.ones_like(val))[0]

    @staticmethod
    def _hess(val, ree):
        """get the hessian of thekernel.

        Warning thos work only because the kernel term are dependent
        of a single rij term, i.e. fij = f(rij)

        Args:
            pos ([type]): [description]
        """

        gval = grad(val,
                    ree,
                    grad_outputs=torch.ones_like(val),
                    create_graph=True)[0]

        hval = grad(gval, ree, grad_outputs=torch.ones_like(gval))[0]

        return hval, gval



from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow
from qmctorch.utils import set_torch_double_precision
from qmctorch.wavefunction.jastrows.elec_elec.kernels import FullyConnectedJastrowKernel

import numpy as np
import torch
import unittest

torch.set_default_tensor_type(torch.DoubleTensor)


class TestSlaterJastrowElectronCusp(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(101)
        np.random.seed(101)

        set_torch_double_precision()

        # molecule
        mol = Molecule(
            atom='He 0.5 0 0; He -0.5 0 0',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g',
            redo_scf=True)

        self.wf = SlaterJastrow(mol,
                                jastrow_kernel=FullyConnectedJastrowKernel,
                                kinetic='jacobi',
                                include_all_mo=True,
                                configs='ground_state')

        self.nbatch = 100

    def test_ee_cusp(self):

        import matplotlib.pyplot as plt
        pos_x = torch.Tensor(np.random.rand(
            self.nbatch,  self.wf.nelec, 3))
        x = torch.linspace(0, 2, self.nbatch)
        pos_x[:, 0, :] = torch.as_tensor([0., 0., 0.]) + 1E-6
        pos_x[:, 1, 0] = 0.
        pos_x[:, 1, 1] = 0.
        pos_x[:, 1, 2] = x

        pos_x[:, 2, :] = 0.5*torch.as_tensor([1., 1., 1.])
        pos_x[:, 3, :] = -0.5*torch.as_tensor([1., 1., 1.])

        pos_x = pos_x.reshape(self.nbatch, self.wf.nelec*3)
        pos_x.requires_grad = True

        x = x.detach().numpy()
        j = self.wf.jastrow(pos_x).detach().numpy()
        plt.plot(x, j)
        plt.show()

        dx = x[1]-x[0]
        dj = (j[1:]-j[0:-1])/dx

        plt.plot(x[:-1], dj/j[:-1])
        plt.show()

        epot = self.wf.electronic_potential(pos_x).detach().numpy()
        ekin = self.wf.kinetic_energy_jacobi(pos_x).detach().numpy()
        eloc = self.wf.local_energy(pos_x).detach().numpy()
        plt.plot(x, epot)
        plt.plot(x, ekin)
        plt.plot(x, eloc)
        plt.show()


if __name__ == "__main__":
    # unittest.main()
    t = TestSlaterJastrowElectronCusp()
    t.setUp()
    t.test_ee_cusp()



import torch


def btrace(M):
    """Computes the trace of batched matrices

    Args:
        M (torch.tensor): matrices of size (Na, Nb, ... Nx, N, N)

    Returns:
        torch.tensor: trace of matrices (Na, Nb, ... Nx)

    Example:
        >>> m = torch.rand(100,5,5)
        >>> tr = btrace(m)
    """
    return torch.diagonal(M, dim1=-2, dim2=-1).sum(-1)


def bproj(M, P):
    """Project batched marices using P^T M P

    Args:
        M (torch.tensor): batched matrices size (..., N, M)
        P (torch.tensor): Porjectors size (..., N, M)

    Returns:
        torch.tensor: Projected matrices
    """
    return P.transpose(1, 2) @ M @ P


def bdet2(M):
    """Computes the determinant of batched 2x2 matrices

    Args:
        M (torch.tensor): input matrices

    Returns:
        torch.tensor: determinants of the matrices
    """

    return M[..., 0, 0] * M[..., 1, 1] - M[..., 0, 1] * M[..., 1, 0]


class BatchDeterminant(torch.autograd.Function):

    @staticmethod
    def forward(ctx, input):

        # LUP decompose the matrices
        inp_lu, pivots = input.lu()
        perm, inpl, inpu = torch.lu_unpack(inp_lu, pivots)

        # get the number of permuations
        s = (pivots != torch.as_tensor(
            range(1, input.shape[1]+1)).int()).sum(1).type(torch.get_default_dtype())

        # get the prod of the diag of U
        d = torch.diagonal(inpu, dim1=-2, dim2=-1).prod(1)

        # assemble
        det = ((-1)**s * d)
        ctx.save_for_backward(input, det)

        return det

    @staticmethod
    def backward(ctx, grad_output):
        '''using jaobi's formula
            d det(A) / d A_{ij} = adj^T(A)_{ij}
        using the adjunct formula
            d det(A) / d A_{ij} = ( (det(A) A^{-1})^T )_{ij}
        '''
        input, det = ctx.saved_tensors
        return (grad_output * det).view(-1, 1, 1) * torch.inverse(input).transpose(1, 2)



import unittest

import numpy as np
import torch
import torch.optim as optim

from qmctorch.sampler import Metropolis
from qmctorch.solver import Solver
from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow
from qmctorch.wavefunction.jastrows.elec_elec.kernels import FullyConnectedJastrowKernel
from qmctorch.utils import set_torch_double_precision


class TestLiH(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(0)
        np.random.seed(0)
        set_torch_double_precision()

        # molecule
        self.mol = Molecule(
            atom='Li 0 0 0; H 0 0 3.015',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g')

        # wave function
        self.wf = SlaterJastrow(self.mol, kinetic='jacobi',
                                jastrow_kernel=FullyConnectedJastrowKernel,
                                configs='single(2,2)',
                                include_all_mo=False)

        # sampler
        self.sampler = Metropolis(
            nwalkers=500,
            nstep=200,
            step_size=0.05,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('normal'),
            move={
                'type': 'all-elec',
                'proba': 'normal'})

        # optimizer
        self.opt = optim.Adam(self.wf.parameters(), lr=0.01)

        # solver
        self.solver = Solver(wf=self.wf, sampler=self.sampler,
                                          optimizer=self.opt)

        # artificial pos
        self.nbatch = 10
        self.pos = torch.as_tensor(np.random.rand(
            self.nbatch, self.wf.nelec*3))
        self.pos.requires_grad = True

    def test_0_wavefunction(self):

        eauto = self.wf.kinetic_energy_autograd(self.pos)
        ejac = self.wf.kinetic_energy_jacobi(self.pos)
        print(torch.stack([eauto, ejac], axis=1).squeeze())
        assert torch.allclose(
            eauto.data, ejac.data, rtol=1E-4, atol=1E-4)

    def test1_single_point(self):

        # sample and compute observables
        obs = self.solver.single_point()
        e, v = obs.energy, obs.variance

    # def test2_wf_opt_grad_auto(self):
    #     self.solver.sampler = self.sampler

    #     self.solver.configure(track=['local_energy'],
    #                           loss='energy', grad='auto')
    #     obs = self.solver.run(5)

    def test3_wf_opt_grad_manual(self):
        self.solver.sampler = self.sampler

        self.solver.configure(track=['local_energy'],
                              loss='energy', grad='manual')
        obs = self.solver.run(5)


if __name__ == "__main__":
    unittest.main()
    # t = TestLiH()
    # t.setUp()
    # t.test1_single_point()
    # t.test3_wf_opt_grad_manual()



import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm

from .stat_utils import (blocking, correlation_coefficient,
                         fit_correlation_coefficient,
                         integrated_autocorrelation_time)


def plot_energy(local_energy, e0=None, show_variance=False):
    """Plot the evolution of the energy

    Args:
        local_energy (np.ndarray): local energies along the trajectory
        e0 (float, optional): Target value for the energy. Defaults to None.
        show_variance (bool, optional): show the variance if True. Defaults to False.
    """

    fig = plt.figure()
    ax = fig.add_subplot(111)

    n = len(local_energy)
    epoch = np.arange(n)

    # get the variance
    energy = np.array([np.mean(e) for e in local_energy])
    variance = np.array([np.var(e) for e in local_energy])

    # plot
    ax.fill_between(epoch, energy - variance, energy +
                    variance, alpha=0.5, color='#4298f4')
    ax.plot(epoch, energy, color='#144477')
    if e0 is not None:
        ax.axhline(e0, color='black', linestyle='--')

    ax.grid()
    ax.set_xlabel('Number of epoch')
    ax.set_ylabel('Energy', color='black')

    if show_variance:
        ax2 = ax.twinx()
        ax2.plot(epoch, variance, color='blue')
        ax2.set_ylabel('variance', color='blue')
        ax2.tick_params(axis='y', labelcolor='blue')
        fig.tight_layout()

    plt.show()


def plot_data(observable, obsname):
    """Plot the evolution a given data

    Args:
        obs_dict (SimpleNamespace): namespace of observable
        obsname (str): name (key) of the desired observable
    """

    fig = plt.figure()
    ax = fig.add_subplot(111)

    data = np.array(observable.__dict__[obsname]).squeeze()
    epoch = np.arange(len(data))
    ax.plot(epoch, data, color='#144477')

    plt.show()


def plot_walkers_traj(eloc, walkers='mean'):
    """Plot the trajectory of all the individual walkers

    Args:
        obs (SimpleNamespace): Namespace of the observables
        walkers (int, str, optional): all, mean or index of a given walker Defaults to 'all'
    """
    nstep, nwalkers = eloc.shape
    celoc = np.cumsum(eloc, axis=0).T
    celoc /= np.arange(1, nstep + 1)

    if walkers is not None:

        if walkers == 'all':
            plt.plot(eloc, 'o', alpha=1 / nwalkers, c='grey')
            cmap = cm.hot(np.linspace(0, 1, nwalkers))
            for i in range(nwalkers):
                plt.plot(celoc.T[:, i], color=cmap[i])

        elif walkers == 'mean':
            plt.plot(eloc, 'o', alpha=1 / nwalkers, c='grey')
            emean = np.mean(celoc.T, axis=1)
            emin = emean.min()
            emax = emean.max()
            delta = emax-emin
            plt.plot(emean, linewidth=5)
            plt.ylim(emin-0.25*delta,emax+0.25*delta)
        else:
            raise ValueError('walkers argument must be all or mean')
        
        plt.grid()
        plt.xlabel('Monte Carlo Steps')
        plt.ylabel('Energy (Hartree)')

    plt.show()


def plot_correlation_coefficient(eloc, size_max=100):
    """Plot the correlation coefficient of the local energy
       and fit the curve to an exp to extract the correlation time.

    Args:
        eloc (np.ndarray): values of the local energy (Nstep, Nwalk)
        size_max (int, optional): maximu number of MC step to consider.Defaults to 100.

    Returns:
        np.ndarray, float: correlation coefficients (size_max, Nwalkers), correlation time
    """

    rho = correlation_coefficient(eloc)

    tau_fit, fitted = fit_correlation_coefficient(
        rho.mean(1)[:size_max])

    plt.plot(rho, alpha=0.25)
    plt.plot(rho.mean(1), linewidth=3, c='black')
    plt.plot(fitted, '--', c='grey')
    plt.xlim([0, size_max])
    plt.ylim([-0.25, 1.5])
    plt.xlabel('MC steps')
    plt.ylabel('Correlation coefficient')
    plt.text(0.5*size_max, 1.05, 'tau=%1.3f' %
             tau_fit, {'color': 'black',  'fontsize': 15})
    plt.grid()
    plt.show()

    return rho, tau_fit


def plot_integrated_autocorrelation_time(eloc, rho=None, size_max=100, C=5):
    """compute/plot the integrated autocorrelation time

    Args:
        eloc (np.ndarray, optional): local energy values (Nstep, Nwalkers)
        rho (np.ndarray, optional): Correlation coefficient. Defaults to None.
        size_max (int, optional): maximu number of MC step to consider.Defaults to 100.
        C (int, optional): [description]. Defaults to 5.
    """

    if rho is None:
        rho = correlation_coefficient(eloc)

    tau = integrated_autocorrelation_time(rho, size_max)

    tc, idx_tc = [], []
    idx = np.arange(1, size_max)
    for iw in range(eloc.shape[1]):

        t = tau[:, iw]
        if len(t[C*t <= idx]) > 0:

            tval = t[C*t <= idx][0]
            ii = np.where(t == tval)[0][0]

            tc.append(tval)
            idx_tc.append(ii)

    plt.plot(tau, alpha=0.25)
    tm = tau.mean(1)
    plt.plot(tm, c='black')
    plt.plot(idx/C, '--', c='grey')

    plt.plot(idx_tc, tc, 'o', alpha=0.25)
    tt = tm[tm*C <= idx][0]
    ii = np.where(tm == tt)[0][0]
    plt.plot(ii, tt, 'o')

    plt.grid()
    plt.xlabel('MC step')
    plt.ylabel('IAC')
    plt.show()


def plot_blocking_energy(eloc, block_size, walkers='mean'):
    """Plot the blocked energy values

    Args:
        eloc (np.ndarray): values of the local energies
        block_size (int): size of the block
        walkers (str, optional): which walkers to plot (mean, all, index or list). Defaults to 'mean'.

    Raises:
        ValueError: [description]
    """
    eb = blocking(eloc, block_size, expand=True)
    if walkers == 'all':
        plt.plot(eloc)
        plt.plot(eb)

    elif walkers == 'mean':
        plt.plot(eloc.mean(1))
        plt.plot(eb.mean(1))

    elif walkers.__class__.__name__ in ['int', 'list']:
        plt.plot(eloc[:, walkers])
        plt.plot(eb[:, walkers])

    else:
        raise ValueError('walkers ', walkers, ' not recognized')

    plt.grid()
    plt.xlabel('MC steps')
    plt.ylabel('Energy')
    plt.show()

    return blocking(eloc, block_size, expand=False)


def plot_correlation_time(eloc):
    """Plot the blocking thingy

    Args:
        eloc (np.array): values of the local energy
    """

    nstep, nwalkers = eloc.shape
    max_block_size = nstep // 2

    var = np.std(eloc, axis=0)

    evar = []
    for size in range(1, max_block_size):
        eb = blocking(eloc, size)
        evar.append(np.std(eb, axis=0) * size / var)

    plt.plot(np.array(evar))
    plt.xlabel('Blocking size')
    plt.ylabel('Correlation steps')
    plt.show()


def plot_block(eloc):
    """Plot the blocking thingy

    Args:
        eloc (np.array): values of the local energy
    """

    nstep, nwalkers = eloc.shape
    max_block_size = nstep // 2

    evar = []
    for size in range(1, max_block_size):
        eb = blocking(eloc, size)
        nblock = eb.shape[0]
        evar.append(np.sqrt(np.var(eb, axis=0) / (nblock - 1)))

    plt.plot(np.array(evar))
    plt.show()



import unittest
import torch
import numpy as np
from types import SimpleNamespace
from qmctorch.wavefunction.orbitals.norm_orbital import atomic_orbital_norm


class TestAtomicOrbitalNorm(unittest.TestCase):

    def test_sph_sto(self):

        basis = SimpleNamespace()
        basis.harmonics_type = 'sph'
        basis.radial_type = 'sto'
        basis.bas_n = torch.as_tensor([0, 1, 2])
        basis.bas_exp = torch.rand(3)

        atomic_orbital_norm(basis)

    def test_sph_gto(self):

        basis = SimpleNamespace()
        basis.harmonics_type = 'sph'
        basis.radial_type = 'gto'
        basis.bas_n = torch.as_tensor([0, 1, 2])
        basis.bas_exp = torch.rand(3)

        atomic_orbital_norm(basis)

    def test_cart_sto(self):

        basis = SimpleNamespace()
        basis.harmonics_type = 'cart'
        basis.radial_type = 'sto'
        basis.bas_exp = np.random.rand(4)
        basis.bas_kx = np.array([0, 0, 0, 1])
        basis.bas_ky = np.array([0, 1, 0, 0])
        basis.bas_kz = np.array([0, 0, 1, 0])
        basis.bas_kr = np.array([0, 0, 0, 0])

        atomic_orbital_norm(basis)

    def test_cart_gto(self):

        basis = SimpleNamespace()
        basis.harmonics_type = 'cart'
        basis.radial_type = 'gto'
        basis.bas_exp = np.random.rand(4)
        basis.bas_kx = np.array([0, 0, 0, 1])
        basis.bas_ky = np.array([0, 1, 0, 0])
        basis.bas_kz = np.array([0, 0, 1, 0])
        basis.bas_kr = np.array([0, 0, 0, 0])

        atomic_orbital_norm(basis)



import numpy as np


class CalculatorBase:
    def __init__(self, atoms, atom_coords, basis, scf, units, molname, calcname, savefile):

        self.atoms = atoms
        self.atom_coords = atom_coords
        self.basis_name = basis
        self.scf = scf
        self.units = units
        self.molname = molname
        self.calcname = calcname
        self.savefile = savefile

    def run(self):
        raise NotImplementedError(
            'Implement a run method in your calculator.')

    def save_data(self):
        raise NotImplementedError(
            'Implement a save_data method in your calculator.')

    @staticmethod
    def normalize_columns(mat):
        """Normalize a matrix column-wise.

        Arguments:
            mat {np.ndarray} -- the matrix to be normalized

        Returns:
            np.ndarray -- normalized matrix
        """
        norm = np.sqrt((mat**2).sum(0))
        return mat / norm




import unittest
import torch

from torch.autograd import Variable, grad, gradcheck

from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow
from ...path_utils import PATH_TEST

torch.set_default_tensor_type(torch.DoubleTensor)


def hess(out, pos):
    # compute the jacobian
    z = Variable(torch.ones(out.shape))
    jacob = grad(out, pos,
                 grad_outputs=z,
                 only_inputs=True,
                 create_graph=True)[0]

    # compute the diagonal element of the Hessian
    z = Variable(torch.ones(jacob.shape[0]))
    hess = torch.zeros(jacob.shape)

    for idim in range(jacob.shape[1]):

        tmp = grad(jacob[:, idim], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, idim] = tmp[:, idim]

    return hess


def hess_mixed_terms(out, pos):

    # compute the jacobian
    z = Variable(torch.ones(out.shape))
    jacob = grad(out, pos,
                 grad_outputs=z,
                 only_inputs=True,
                 create_graph=True)[0]

    # compute the diagonal element of the Hessian
    z = Variable(torch.ones(jacob.shape[0]))
    hess = torch.zeros(jacob.shape)
    nelec = pos.shape[1]//3
    k = 0

    for ielec in range(nelec):

        ix = ielec*3
        tmp = grad(jacob[:, ix], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, k] = tmp[:, ix+1]
        k = k + 1
        hess[:, k] = tmp[:, ix+2]
        k = k + 1

        iy = ielec*3 + 1
        tmp = grad(jacob[:, iy], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, k] = tmp[:, iy+1]
        k = k + 1

    return hess


class TestAOderivativesADF(unittest.TestCase):

    def setUp(self):

        # define the molecule
        path_hdf5 = PATH_TEST / 'hdf5/C_adf_dzp.hdf5'
        self.mol = Molecule(load=path_hdf5)

        # define the wave function
        self.wf = SlaterJastrow(self.mol, include_all_mo=True)

        # define the grid points
        npts = 11
        self.pos = torch.rand(npts, self.mol.nelec * 3)
        self.pos = Variable(self.pos)
        self.pos.requires_grad = True

    def test_ao_deriv(self):

        ao = self.wf.ao(self.pos)
        dao = self.wf.ao(self.pos, derivative=1)
        dao_grad = grad(
            ao, self.pos, grad_outputs=torch.ones_like(ao))[0]

        gradcheck(self.wf.ao, self.pos)
        assert(torch.allclose(dao.sum(), dao_grad.sum()))

    def test_ao_grad_sum(self):

        ao = self.wf.ao(self.pos)
        dao_sum = self.wf.ao(self.pos, derivative=1, sum_grad=True)
        dao = self.wf.ao(self.pos, derivative=1, sum_grad=False)

        assert(torch.allclose(dao_sum, dao.sum(-1)))

    def test_ao_hess(self):

        ao = self.wf.ao(self.pos)
        d2ao = self.wf.ao(self.pos, derivative=2)
        d2ao_grad = hess(ao, self.pos)
        assert(torch.allclose(d2ao.sum(), d2ao_grad.sum()))

    def test_ao_hess_sum(self):

        ao = self.wf.ao(self.pos)
        d2ao_sum = self.wf.ao(self.pos, derivative=2, sum_hess=True)
        d2ao = self.wf.ao(self.pos, derivative=2, sum_hess=False)
        assert(torch.allclose(d2ao_sum, d2ao.sum(-1)))

    def test_ao_mixed_der(self):
        ao = self.wf.ao(self.pos)
        d2ao = self.wf.ao(self.pos, derivative=3)
        d2ao_auto = hess_mixed_terms(ao, self.pos)
        assert(torch.allclose(d2ao.sum(), d2ao_auto.sum()))


if __name__ == "__main__":
    # unittest.main()

    t = TestAOderivativesADF()
    t.setUp()
    t.test_ao_deriv()
    t.test_ao_hess()



"""Utils module API."""

from .algebra_utils import bdet2, bproj, btrace
from .hdf5_utils import (add_group_attr, dump_to_hdf5, load_from_hdf5,
                         register_extra_attributes, bytes2str)
from .interpolate import (InterpolateAtomicOrbitals,
                          InterpolateMolecularOrbitals)
from .plot_data import (plot_block, plot_blocking_energy,
                        plot_correlation_coefficient, plot_correlation_time,
                        plot_data, plot_energy,
                        plot_integrated_autocorrelation_time,
                        plot_walkers_traj)
from .stat_utils import (blocking, correlation_coefficient,
                         integrated_autocorrelation_time)
from .torch_utils import (DataSet, Loss, OrthoReg, fast_power,
                          set_torch_double_precision,
                          set_torch_single_precision,
                          diagonal_hessian, gradients)

__all__ = ['plot_energy', 'plot_data', 'plot_block',
           'plot_walkers_traj',
           'plot_correlation_time',
           'plot_autocorrelation',
           'set_torch_double_precision',
           'set_torch_single_precision',
           'DataSet', 'Loss', 'OrthoReg',
           'dump_to_hdf5', 'load_from_hdf5',
           'bytes2str',
           'register_extra_attributes',
           'fast_power',
           'InterpolateMolecularOrbitals',
           'InterpolateAtomicOrbitals',
           'btrace', 'bdet2', 'bproj',
           'diagonal_hessian', 'gradients']



import torch
import operator

from .slater_jastrow_base import SlaterJastrowBase
from .jastrows.elec_elec.kernels.pade_jastrow_kernel import PadeJastrowKernel
from .jastrows.elec_elec.jastrow_factor_electron_electron import JastrowFactorElectronElectron


class SlaterOrbitalDependentJastrow(SlaterJastrowBase):

    def __init__(self, mol,
                 configs='ground_state',
                 kinetic='jacobi',
                 jastrow_kernel=PadeJastrowKernel,
                 jastrow_kernel_kwargs={},
                 cuda=False,
                 include_all_mo=True):
        """Slater Jastrow Wave function with an orbital dependent Electron-Electron Jastrow Factor

        .. math::
            \\Psi(R_{at}, r) = \\sum_n c_n D^\\uparrow_n(r^\\uparrow)D^\\downarrow_n(r^\\downarrow)

        where each molecular orbital of the determinants is multiplied with a different electron-electron Jastrow

        .. math::
            \\phi_i(r) \\rightarrow J_i(r) \\phi_i(r) 

        Args:
            mol (Molecule): a QMCTorch molecule object
            configs (str, optional): defines the CI configurations to be used. Defaults to 'ground_state'.
                - ground_state : only the ground state determinant in the wave function
                - single(n,m) : only single excitation with n electrons and m orbitals 
                - single_double(n,m) : single and double excitation with n electrons and m orbitals
                - cas(n, m) : all possible configuration using n eletrons and m orbitals                   
            kinetic (str, optional): method to compute the kinetic energy. Defaults to 'jacobi'.
                - jacobi : use the Jacobi formula to compute the kinetic energy 
                - auto : use automatic differentiation to compute the kinetic energy
            jastrow_kernel (JastrowKernelBase, optional) : Class that computes the jastrow kernels
            jastrow_kernel_kwargs (dict, optional) : keyword arguments for the jastrow kernel contructor
            cuda (bool, optional): turns GPU ON/OFF  Defaults to False.
            include_all_mo (bool, optional): include either all molecular orbitals or only the ones that are
                                             popualted in the configs. Defaults to False
        Examples::
            >>> from qmctorch.scf import Molecule
            >>> from qmctorch.wavefunction import SlaterOrbitalDependentJastrow
            >>> mol = Molecule('h2o.xyz', calculator='adf', basis = 'dzp')
            >>> wf = SlaterOrbitalDependentJastrow(mol, configs='cas(2,2)')
        """

        if jastrow_kernel is None:
            raise ValueError(
                'Orbital dependent Jastrow factor requires a valid jastrow kernel.')

        super().__init__(mol, configs, kinetic, cuda, include_all_mo)
        self.use_jastrow = True

        self.jastrow = JastrowFactorElectronElectron(
            self.mol.nup, self.mol.ndown, jastrow_kernel,
            kernel_kwargs=jastrow_kernel_kwargs,
            orbital_dependent_kernel=True,
            number_of_orbitals=self.nmo_opt,
            cuda=self.cuda)

        if self.cuda:
            self.jastrow = self.jastrow.to(self.device)

        self.jastrow_type = jastrow_kernel.__class__.__name__

        self.log_data()

    def ordered_jastrow(self, pos, derivative=0, sum_grad=True):
        """Returns the value of the jastrow with the correct dimensions

        Args:
            pos (torch.tensor): Positions of the electrons
                                  Size : Nbatch, Nelec x Ndim
            derivative (int, optional): order of the derivative (0,1,2,).
                            Defaults to 0.
            sum_grad (bool, optional): Return the sum_grad (i.e. the sum of
                                       the derivatives) or the individual
                                       terms. Defaults to True.
                                       False only for derivative=1

        Returns:
            torch.tensor: value of the jastrow parameter for all confs
                          Nbatch, Nelec, Nmo (sum_grad = True)
                          Nbatch, Nelec, Nmo, Ndim (sum_grad = False)
        """
        jast_vals = self.jastrow(pos, derivative, sum_grad)

        def permute(vals):
            """transpose the data depending on the number of dim."""
            if vals.ndim == 3:
                return vals.permute(1, 2, 0)
            elif vals.ndim == 4:
                return vals.permute(1, 3, 0, 2)

        if isinstance(jast_vals, tuple):
            return tuple([permute(v) for v in jast_vals])
        else:
            return permute(jast_vals)

    def forward(self, x, ao=None):
        """computes the value of the wave function for the sampling points

        .. math::
            \\Psi(R) = \\sum_{n} c_n D^{u}_n(r^u) \\times D^{d}_n(r^d)

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)
            ao (torch.tensor, optional): values of the atomic orbitals (Nbatch, Nelec, Nao)

        Returns:
            torch.tensor: values of the wave functions at each sampling point (Nbatch, 1)

        Examples::
            >>> mol = Molecule('h2.xyz', calculator='adf', basis = 'dzp')
            >>> wf = SlaterJastrow(mol, configs='cas(2,2)')
            >>> pos = torch.rand(500,6)
            >>> vals = wf(pos)
        """

        # compute the jastrow from the pos
        J = self.ordered_jastrow(x)

        # atomic orbital
        if ao is None:
            x = self.ao(x)
        else:
            x = ao

        # molecular orbitals
        x = self.mo_scf(x)

        # mix the mos
        x = self.mo(x)

        # jastrow for each orbital
        x = J * x

        # pool the mos
        x = self.pool(x)

        # compute the CI and return
        return self.fc(x)

    def ao2mo(self, ao):
        return self.mo(self.mo_scf(ao))

    def ao2cmo(self, ao, jastrow):
        return jastrow * self.mo(self.mo_scf(ao))

    def pos2mo(self, x, derivative=0, sum_grad=True):
        """Compute the uncorrelated MOs from the positions."""

        ao = self.ao(x, derivative=derivative, sum_grad=sum_grad)
        if sum_grad:
            return self.ao2mo(ao)
        else:
            return self.ao2mo(ao.transpose(2, 3)).transpose(2, 3)

    def pos2cmo(self, x, derivative=0, sum_grad=True):
        """Get the values of correlated MOs

        Arguments:
            x {torch.tensor} -- positions of the electrons [nbatch, nelec*ndim]


        Returns:
            torch.tensor -- MO matrix [nbatch, nelec, nmo]
        """
        if derivative == 0:
            mo = self.pos2mo(x)
            jast = self.ordered_jastrow(x)
            return jast * mo

        elif derivative == 1:

            mo = self.pos2mo(x)
            dmo = self.pos2mo(x, derivative=1, sum_grad=sum_grad)

            jast = self.ordered_jastrow(x)
            djast = self.ordered_jastrow(
                x, derivative=1, sum_grad=sum_grad)

            if sum_grad:
                return mo * djast.sum(1).unsqueeze(1) + jast * dmo
            else:
                return mo.unsqueeze(-1) * djast.sum(1).unsqueeze(1) + jast.unsqueeze(-1) * dmo

        elif derivative == 2:

            # atomic orbital
            ao, dao, d2ao = self.ao(x, derivative=[0, 1, 2])

            # bare molecular orbitals
            mo = self.ao2mo(ao)
            dmo = self.ao2mo(dao.transpose(2, 3)).transpose(2, 3)
            d2mo = self.ao2mo(d2ao)

            # jastrows
            jast, djast, d2jast = self.ordered_jastrow(x,
                                                       derivative=[
                                                           0, 1, 2],
                                                       sum_grad=False)
            # terms of the kin op
            jast_d2mo = d2mo * jast
            djast_dmo = (djast * dmo).sum(-1)
            d2jast_mo = d2jast.sum(1).unsqueeze(1) * mo

            # assemble kin op
            return jast_d2mo + 2 * djast_dmo + d2jast_mo

    def kinetic_energy_jacobi(self, x,  **kwargs):
        """Compute the value of the kinetic enery using the Jacobi Formula.
        C. Filippi, Simple Formalism for Efficient Derivatives .

        .. math::

             \\frac{K(R)}{\\Psi(R)} = Tr(A^{-1} B_{kin})

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)

        Returns:
            torch.tensor: values of the kinetic energy at each sampling points
        """

        # get the matrix of correlated orbitals for all elec
        cmo = self.pos2cmo(x)

        # compute the value of the slater det
        slater_dets = self.pool(cmo)

        # compute  \Delta A (A = matrix of the correlated MO)
        bhess = self.pos2cmo(x, 2)

        # compute ( tr(A_u^-1\Delta A_u) + tr(A_d^-1\Delta A_d) )
        hess = self.pool.operator(cmo, bhess)

        # compute \grad A
        bgrad = self.get_gradient_operator(x)

        # compute (tr(A_u^-1\nabla A_u) * tr(A_d^-1\nabla A_d))
        grad = self.pool.operator(cmo, bgrad, op=None)
        grad2 = self.pool.operator(cmo, bgrad, op_squared=True)

        # assemble the total kinetic values
        kin = - 0.5 * (hess
                       + operator.add(*[(g**2).sum(0) for g in grad])
                       - grad2.sum(0)
                       + 2 * operator.mul(*grad).sum(0))

        # assemble
        return self.fc(kin * slater_dets) / self.fc(slater_dets)

    def gradients_jacobi(self, x, sum_grad=True, pdf=False):
        """Computes the gradients of the wf using Jacobi's Formula

        Args:
            x ([type]): [description]
        """

        if pdf:
            raise NotImplementedError(
                'Gradients of the pdf not implemented for ', self.__name__)

        # get the CMO matrix
        cmo = self.pos2cmo(x)

        # get the grad of the wf
        if sum_grad:
            # bgrad = self.pos2cmo(x, derivative=1)
            bgrad = self.get_gradient_operator(x).sum(0)
        else:
            bgrad = self.get_gradient_operator(x)

        # compute the value of the grad using trace trick
        grad = self.pool.operator(cmo, bgrad, op=operator.add)

        # compute the total wf
        psi = self.pool(cmo)

        out = self.fc(grad * psi)
        out = out.transpose(0, 1)

        # assemble
        return out

    def get_hessian_operator(self, x):
        """Compute the Bkin matrix

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)
            mo (torch.tensor, optional): precomputed values of the MOs

        Returns:
            torch.tensor: matrix of the kinetic operator
        """
        mo = self.pos2mo(x)
        dmo = self.pos2mo(x, derivative=1, sum_grad=False)
        d2mo = self.pos2mo(x, derivative=2)

        jast = self.ordered_jastrow(x)
        djast = self.ordered_jastrow(
            x, derivative=1, sum_grad=False)
        d2jast = self.ordered_jastrow(x, derivative=2)

        # \Delta_n J * MO
        d2jast_mo = d2jast.permute(1, 0, 2).unsqueeze(2) * mo

        # stride d2mo
        eye = torch.eye(self.nelec).to(self.device)
        d2mo = d2mo.unsqueeze(2) * eye.unsqueeze(-1)

        # reshape d2mo to nelec, nbatch, nelec, nmo
        d2mo = d2mo.permute(1, 0, 2, 3)

        # \Delta_n MO * J
        d2mo_jast = d2mo * jast.repeat(1, self.nelec, 1)

        # reformat to have Ndim, Nbatch, Nelec, Nmo
        dmo = dmo.permute(3, 0, 1, 2)

        # stride
        eye = torch.eye(self.nelec).to(self.device)
        dmo = dmo.unsqueeze(2) * eye.unsqueeze(-1)

        # reorder to have Nelec, Ndim, Nbatch, Nelec, Nmo
        dmo = dmo.permute(2, 0, 1, 3, 4)

        # reshape djast to Nelec, Ndim, Nbatch, 1, Nmo
        djast = djast.permute(1, 3, 0, 2).unsqueeze(-2)

        # \nabla jast \nabla mo
        djast_dmo = (djast * dmo)

        # sum over ndim -> Nelec, Nbatch, Nelec, Nmo
        djast_dmo = djast_dmo.sum(1)

        return d2mo_jast + d2jast_mo + 2*djast_dmo

    def get_gradient_operator(self, x):
        """Compute the gradient operator

        Args:
            x ([type]): [description]
            ao ([type]): [description]
            dao ([type]): [description]
        """

        mo = self.pos2mo(x)
        dmo = self.pos2mo(x, derivative=1, sum_grad=False)

        jast = self.ordered_jastrow(x)
        djast = self.ordered_jastrow(x, derivative=1, sum_grad=False)

        # reformat to have Nelec, Ndim, Nbatch, 1, Nmo
        djast = djast.permute(1, 3, 0, 2).unsqueeze(-2)

        # reformat to have Ndim, Nbatch, Nelec, Nmo
        dmo = dmo.permute(3, 0, 1, 2)

        # stride the tensor
        eye = torch.eye(self.nelec).to(self.device)
        dmo = dmo.unsqueeze(2) * eye.unsqueeze(-1)

        # reorder to have Nelec, Ndim, Nbatch, Nelec, Nmo
        dmo = dmo.permute(2, 0, 1, 3, 4)

        # assemble the derivative
        out = (mo * djast + dmo * jast)

        # collapse the first two dimensions
        out = out.reshape(-1, *(out.shape[2:]))
        return out



import unittest

import matplotlib.pyplot as plt
import numpy as np
import torch
from pyscf import gto
from torch.autograd import Variable

from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow

__PLOT__ = False


class TestAOvaluesPyscf(unittest.TestCase):

    def setUp(self):

        # define the molecule
        at = 'C 0 0 0'
        basis = 'dzp'
        self.mol = Molecule(atom=at,
                            calculator='pyscf',
                            basis=basis,
                            unit='bohr')

        self.m = gto.M(atom=at, basis=basis, unit='bohr')

        # define the wave function
        self.wf = SlaterJastrow(self.mol)

        self.pos = torch.zeros(100, self.mol.nelec * 3)

        self.pos[:, 0] = torch.linspace(-5, 5, 100)
        self.pos[:, 1] = torch.linspace(-5, 5, 100)
        self.pos[:, 2] = torch.linspace(-5, 5, 100)

        self.pos = Variable(self.pos)
        self.pos.requires_grad = True

        self.x = self.pos[:, 0].detach().numpy()

    def test_ao(self):

        nzlm = np.linalg.norm(self.m.cart2sph_coeff(), axis=1)

        aovals = self.wf.ao(self.pos).detach().numpy()/nzlm
        aovals_ref = self.m.eval_ao('GTOval_cart',
                                    self.pos.detach().numpy()[:, :3])

        for iorb in range(self.mol.basis.nao):

            if __PLOT__:

                plt.plot(self.x, aovals[:, 0, iorb])
                plt.plot(self.x, aovals_ref[:, iorb])
                plt.show()

            assert np.allclose(
                aovals[:, 0, iorb], aovals_ref[:, iorb])

    def test_ao_deriv(self):

        nzlm = np.linalg.norm(self.m.cart2sph_coeff(), axis=1)

        daovals = self.wf.ao(
            self.pos, derivative=1).detach().numpy()/nzlm

        daovals_ref = self.m.eval_gto(
            'GTOval_ip_cart', self.pos.detach().numpy()[:, :3])
        daovals_ref = daovals_ref.sum(0)

        for iorb in range(self.mol.basis.nao):

            if __PLOT__:
                plt.plot(self.x, daovals[:, 0, iorb])
                plt.plot(self.x, daovals_ref[:, iorb])
                plt.show()

            assert np.allclose(
                daovals[:, 0, iorb], daovals_ref[:, iorb])


if __name__ == "__main__":
    unittest.main()



import torch
from torch import nn
from ..distance.electron_electron_distance import ElectronElectronDistance
from .orbital_dependent_jastrow_kernel import OrbitalDependentJastrowKernel


class JastrowFactorElectronElectron(nn.Module):

    def __init__(self, nup, ndown,
                 jastrow_kernel,
                 kernel_kwargs={},
                 orbital_dependent_kernel=False,
                 number_of_orbitals=None,
                 scale=False, scale_factor=0.6,
                 cuda=False):
        """Electron-Electron Jastrow factor.

        .. math::
            J = \\prod_{i<j} \\exp(\\text{Kernel}(r_{ij}))

        Args:
            nup (int): number of spin up electons
            ndow (int): number of spin down electons
            jastrow_kernel (kernel): class of a electron-electron Jastrow kernel
            kernel_kwargs (dict, optional): keyword argument of the kernel. Defaults to {}.
            orbital_dependent_kernel (bool, optional): Make the kernel orbital dependent. Defaults to False.
            number_of_orbitals (int, optional): number of orbitals for orbital dependent kernels. Defaults to None.
            scale (bool, optional): use scaled electron-electron distance. Defaults to False.
            scale_factor (float, optional): scaling factor. Defaults to 0.6.
            cuda (bool, optional): use cuda. Defaults to False.
        """

        super().__init__()

        self.nup = nup
        self.ndown = ndown
        self.nelec = nup + ndown
        self.ndim = 3

        self.cuda = cuda
        self.device = torch.device('cpu')
        if self.cuda:
            self.device = torch.device('cuda')

        self.requires_autograd = True

        # kernel function
        if orbital_dependent_kernel:
            self.jastrow_kernel = OrbitalDependentJastrowKernel(
                nup, ndown, number_of_orbitals, cuda, jastrow_kernel, kernel_kwargs)
        else:
            self.jastrow_kernel = jastrow_kernel(
                nup, ndown, cuda, **kernel_kwargs)
            self.requires_autograd = self.jastrow_kernel.requires_autograd

        # mask to extract the upper diag of the matrices
        self.mask_tri_up, self.index_col, self.index_row = self.get_mask_tri_up()

        # elec-elec distances
        self.edist = ElectronElectronDistance(self.nelec, self.ndim,
                                              scale=scale,
                                              scale_factor=scale_factor)

    def get_mask_tri_up(self):
        r"""Get the mask to select the triangular up matrix

        Returns:
            torch.tensor: mask of the tri up matrix
        """
        mask = torch.zeros(self.nelec, self.nelec).type(
            torch.bool).to(self.device)
        index_col, index_row = [], []
        for i in range(self.nelec-1):
            for j in range(i+1, self.nelec):
                index_row.append(i)
                index_col.append(j)
                mask[i, j] = True

        index_col = torch.LongTensor(index_col).to(self.device)
        index_row = torch.LongTensor(index_row).to(self.device)
        return mask, index_col, index_row

    def extract_tri_up(self, inp):
        r"""extract the upper triangular elements

        Args:
            inp (torch.tensor): input matrices (nbatch, n, n)

        Returns:
            torch.tensor: triangular up element (nbatch, -1)
        """
        nbatch = inp.shape[0]
        return inp.masked_select(self.mask_tri_up).view(nbatch, -1)

    def get_edist_unique(self, pos, derivative=0):
        """Get the unique elements of the electron-electron distance matrix.

        Args:
            pos (torch.tensor): positions of the electrons (Nbatch, Nelec*3)
            derivative(int, optional): order of the derivative

        Returns:
            torch.tensor: unique values of the electron-electron distance matrix
        """

        if derivative == 0:
            return self.extract_tri_up(self.edist(pos))

        elif derivative == 1:
            nbatch = pos.shape[0]
            return self.extract_tri_up(self.edist(
                pos, derivative=1)).view(nbatch, 3, -1)

        elif derivative == 2:
            nbatch = pos.shape[0]
            return self.extract_tri_up(self.edist(
                pos, derivative=2)).view(nbatch, 3, -1)

    def forward(self, pos, derivative=0, sum_grad=True):
        """Compute the Jastrow factors.

        Args:
            pos (torch.tensor): Positions of the electrons
                                  Size : Nbatch, Nelec x Ndim
            derivative (int, optional): order of the derivative (0,1,2,).
                            Defaults to 0.
            sum_grad (bool, optional): Return the sum_grad (i.e. the sum of
                                       the derivatives) or the individual
                                       terms. Defaults to True.
                                       False only for derivative=1

        Returns:
            torch.tensor: value of the jastrow parameter for all confs
                          derivative = 0  (Nmo) x Nbatch x 1
                          derivative = 1  (Nmo) x Nbatch x Nelec (for sum_grad = True)
                          derivative = 1  (Nmo) x Nbatch x Ndim x Nelec (for sum_grad = False)
                          derivative = 2  (Nmo) x Nbatch x Nelec
        """

        size = pos.shape
        assert size[1] == self.nelec * self.ndim

        r = self.get_edist_unique(pos)
        kern_vals = self.jastrow_kernel(r)
        jast = torch.exp(kern_vals.sum(-1)).unsqueeze(-1)

        if derivative == 0:
            return jast

        elif derivative == 1:
            dr = self.get_edist_unique(pos, derivative=1)
            return self.jastrow_factor_derivative(r, dr, jast, sum_grad)

        elif derivative == 2:

            dr = self.get_edist_unique(pos, derivative=1)
            d2r = self.get_edist_unique(pos, derivative=2)

            return self.jastrow_factor_second_derivative(r, dr, d2r, jast)

        elif derivative == [0, 1, 2]:

            dr = self.get_edist_unique(pos, derivative=1)
            d2r = self.get_edist_unique(pos, derivative=2)

            return(jast,
                   self.jastrow_factor_derivative(
                       r, dr, jast, sum_grad),
                   self.jastrow_factor_second_derivative(r, dr, d2r, jast))

    def jastrow_factor_derivative(self, r, dr, jast, sum_grad):
        """Compute the value of the derivative of the Jastrow factor

        Args:
            r (torch.tensor): distance matrix Nbatch x Nelec x Nelec
            dr (torch.tensor): derivative of the distance matrix Nbatch x Nelec x Nelec x 3
            jast (torch.tensor): values of the jastrow elements
                                 Nbatch x Nelec x Nelec
            sum_grad (bool): return the sum of the gradient or the individual components

        Returns:
            torch.tensor: gradient of the jastrow factors
                          Nbatch x Nelec x Ndim
        """

        if sum_grad:

            djast = self.jastrow_kernel.compute_derivative(
                r, dr).sum(-2)
            djast = djast * jast

            # might cause problems with backward cause in place operation
            out_shape = list(djast.shape[:-1]) + [self.nelec]
            out = torch.zeros(out_shape).to(self.device)
            out.index_add_(-1, self.index_row, djast)
            out.index_add_(-1, self.index_col, -djast)

        else:

            djast = self.jastrow_kernel.compute_derivative(
                r, dr)
            djast = djast * jast.unsqueeze(-1)

            # might cause problems with backward cause in place operation
            out_shape = list(djast.shape[:-1]) + [self.nelec]
            out = torch.zeros(out_shape).to(self.device)
            out.index_add_(-1, self.index_row, djast)
            out.index_add_(-1, self.index_col, -djast)

        return out

    def jastrow_factor_second_derivative(self, r, dr, d2r, jast):
        """Compute the value of the pure 2nd derivative of the Jastrow factor

        Args:
            r (torch.tensor): distance matrix Nbatch x Nelec x Nelec
            dr (torch.tensor): derivative of the distance matrix Nbatch x Nelec x Nelec x 3
            d2r (torch.tensor): 2nd derivative of the distance matrix Nbatch x Nelec x Nelec x 3
            jast (torch.tensor): values of the ajstrow elements
                                 Nbatch x Nelec x Nelec

        Returns:
            torch.tensor: diagonal hessian of the jastrow factors
                          Nbatch x Nelec x Ndim
        """

        d2jast = self.jastrow_kernel.compute_second_derivative(
            r, dr, d2r).sum(-2)

        # might cause problems with backward cause in place operation
        hess_shape = list(d2jast.shape[:-1]) + [self.nelec]
        hess_jast = torch.zeros(hess_shape).to(self.device)
        hess_jast.index_add_(-1, self.index_row, d2jast)
        hess_jast.index_add_(-1, self.index_col, d2jast)

        # mixed terms
        djast = self.jastrow_kernel.compute_derivative(r, dr)

        # add partial derivative
        hess_jast = hess_jast + self.partial_derivative(djast)

        return hess_jast * jast

    def partial_derivative(self, djast):
        """Computes the partial derivative

        Args:
            djast (torch.tensor): values of the derivative of the jastrow kernels
        """

        out_shape = list(djast.shape[:-1]) + [self.nelec]
        out = torch.zeros(out_shape).to(self.device)
        out.index_add_(-1, self.index_row, djast)
        out.index_add_(-1, self.index_col, -djast)
        return (out**2).sum(-2)



from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrowBackFlow, SlaterJastrow
from qmctorch.utils import set_torch_double_precision
from qmctorch.wavefunction.orbitals.backflow.kernels import BackFlowKernelInverse

from torch.autograd import grad, Variable

import numpy as np
import torch
import unittest

torch.set_default_tensor_type(torch.DoubleTensor)


def hess(out, pos):
    # compute the jacobian
    z = Variable(torch.ones(out.shape))
    jacob = grad(out, pos,
                 grad_outputs=z,
                 only_inputs=True,
                 create_graph=True)[0]

    # compute the diagonal element of the Hessian
    z = Variable(torch.ones(jacob.shape[0]))
    hess = torch.zeros(jacob.shape)

    for idim in range(jacob.shape[1]):

        tmp = grad(jacob[:, idim], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, idim] = tmp[:, idim]

    return hess


class TestCompareSlaterJastrowBackFlow(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(101)
        np.random.seed(101)

        set_torch_double_precision()

        # molecule
        mol = Molecule(
            atom='Li 0 0 0; H 0 0 3.015',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g',
            redo_scf=True)

        self.wf = SlaterJastrowBackFlow(mol,
                                        kinetic='jacobi',
                                        include_all_mo=True,
                                        configs='single_double(2,2)',
                                        backflow_kernel=BackFlowKernelInverse,
                                        orbital_dependent_backflow=False)

        self.wf.ao.backflow_trans.backflow_kernel.weight.data *= 0.

        self.wf_ref = SlaterJastrow(mol,
                                    kinetic='jacobi',
                                    include_all_mo=True,
                                    configs='single_double(2,2)')

        self.random_fc_weight = torch.rand(self.wf.fc.weight.shape)
        self.wf.fc.weight.data = self.random_fc_weight
        self.wf_ref.fc.weight.data = self.random_fc_weight

        self.nbatch = 5
        self.pos = torch.Tensor(np.random.rand(
            self.nbatch,  self.wf.nelec*3))
        self.pos.requires_grad = True

    def test_forward(self):
        """Check that backflow give same results as normal SlaterJastrow."""
        wf_val = self.wf(self.pos)
        wf_val_ref = self.wf_ref(self.pos)

        assert(torch.allclose(wf_val, wf_val_ref))

    def test_jacobian_mo(self):
        """Check that backflow give same results as normal SlaterJastrow."""

        dmo = self.wf.pos2mo(self.pos, derivative=1)
        dmo_ref = self.wf_ref.pos2mo(self.pos, derivative=1)
        assert(torch.allclose(dmo.sum(0), dmo_ref))

    def test_hess_mo(self):
        """Check that backflow give same results as normal SlaterJastrow."""
        d2ao = self.wf.ao(self.pos, derivative=2, sum_hess=False)
        d2val = self.wf.ao2mo(d2ao)

        d2ao_ref = self.wf_ref.ao(
            self.pos, derivative=2, sum_hess=True)
        d2val_ref = self.wf_ref.ao2mo(d2ao_ref)
        assert(torch.allclose(d2val_ref, d2val.sum(0)))

    def test_grad_wf(self):
        pass

    def test_local_energy(self):

        self.wf.kinetic_energy = self.wf.kinetic_energy_jacobi
        eloc_jac = self.wf.local_energy(self.pos)

        self.wf_ref.kinetic_energy = self.wf_ref.kinetic_energy_jacobi
        eloc_jac_ref = self.wf_ref.local_energy(self.pos)

        assert torch.allclose(
            eloc_jac_ref.data, eloc_jac.data, rtol=1E-4, atol=1E-4)

    def test_kinetic_energy(self):

        ejac_ref = self.wf_ref.kinetic_energy_jacobi(self.pos)
        ejac = self.wf.kinetic_energy_jacobi(self.pos)

        assert torch.allclose(
            ejac_ref.data, ejac.data, rtol=1E-4, atol=1E-4)


if __name__ == "__main__":
    t = TestCompareSlaterJastrowBackFlow()
    t.setUp()
    t.test_jacobian_mo()
    t.test_hess_mo()
    t.test_kinetic_energy()
    t.test_local_energy()
    # unittest.main()
